from __future__ import annotations

import re
import warnings
from pathlib import Path

from pyjelly.integrations.generic.generic_sink import (
    DEFAULT_GRAPH_IDENTIFIER,
    IRI,
    TRIPLE_ARITY,
    BlankNode,
    GenericStatementSink,
    Literal,
    Prefix,
    Quad,
    Triple,
)


class GenericSinkParser:
    _uri_re = re.compile(r"<([^>\s]+)>")  # returns URI
    _bn_re = re.compile(r"_:(\S+)")  # returns blank node identificator
    _literal_re = re.compile(
        r""""([^"]*)"(?:@(\S+)|\^\^<(\S+)>)?"""
    )  # returns lex part of the literal and optional langtag and datatype
    _quoted_triple_re = re.compile(
        r"<<.*?>>"
    )  # matches the quoted triple quotation marks syntax
    _token_quoted_triple_re = re.compile(
        r"<<\s*(.*?)\s*>>"
    )  # returns the quoted triple
    _prefix_re = re.compile(
        r"@prefix\s+(\w+):\s*<([^>]+)>\s*\."
    )  # returns prefix and namespace IRI
    _split_tokens = re.compile(  # matches str to quoted triple/literal, IRI, or BN
        r"""
        <<.+?>>         |
        "[^"]+"(?:@\S+|\^\^\S+)?  |
        <[^>]+>         |
        _:\S+           |
        """,
        re.VERBOSE,
    )

    def __init__(self, sink: GenericStatementSink) -> None:
        self._sink = sink

    def process_term(
        self, term: str, slot: str
    ) -> IRI | BlankNode | Literal | Triple | str:
        """
        Process one term.

        Notes:
            terms are not validated to match RDF spec.

        Args:
            term (str): term to process
            slot (str): data structure to associate with terms

        Raises:
            TypeError: if literal decoded has multiple parts except for lex, i.e.,
                possibly both langtag and datatype are specified.
            TypeError: if fail to parse quoted triple into valid s/p/o.
            TypeError: if term did not match any pattern

        Returns:
            IRI | BlankNode | Literal | Triple | str: processed term

        """
        match_bn = self._bn_re.match(term)
        if match_bn:
            return BlankNode(match_bn.groups()[0])
        match_iri = self._uri_re.match(term)
        if match_iri:
            return IRI(match_iri.groups()[0])
        match_literal = self._literal_re.match(term)
        if match_literal:
            lex, langtag, datatype = match_literal.groups()
            if not lex or (langtag is not None and datatype is not None):
                msg = "invalid literal encountered"
                raise TypeError(msg)
            return Literal(lex, langtag, datatype)

        match_quoted_triple = self._quoted_triple_re.match(term)
        if match_quoted_triple:
            triple_tokens = self.split_statement(term.strip())
            return Triple(
                *(
                    self.process_term(group, slot)
                    for slot, group in zip(Triple._fields, triple_tokens)
                )
            )
        if term == "" and slot == "g":
            return DEFAULT_GRAPH_IDENTIFIER

        msg = "failed to parse input file"
        raise TypeError(msg)

    def find_closing_quotation(self, text: str, start: int) -> int:
        """
        Find the closing >> to a passed one.

        Notes:
            Examines what is inside the current quoted triple.
            Counts nesting levels and reports only the paired
            closing quote.

        Args:
            text (str): the current quoted triple parsed
            start (int): location of the opning quotation

        Returns:
            int: position of the closing quotation mark.

        """
        depth = 1
        pos = start + 2
        while pos < len(text) - 1 and depth > 0:
            if text[pos : pos + 2] == "<<":
                depth += 1
                pos += 2
            elif text[pos : pos + 2] == ">>":
                depth -= 1
                pos += 2
            else:
                pos += 1
        return pos

    def process_quoted(self, statement: str) -> list[str]:
        """
        Process quoted triple or its part.

        Notes:
            Allows having more quoted triples inside.
            Decides whether to:
                - tokenize the current triple, if no more nesting found or a part of
                    quoted triple is given
                - return as is -- if the statement is this only triple,
                    used during splitting statements into s/p/o/g
                - split quoted triple further -- during active term processing,
                    goes until all nesting levels are exhausted and the first
                    if is triggered

        Args:
            statement (str): quoted triple to process or a part of one

        Returns:
            list[str]: list of processed triple terms

        """
        start = statement.find("<<")
        if start == -1:
            split = self.generate_statement_tokens(statement.strip())
        else:
            if start == 0 and self.find_closing_quotation(statement, start) == len(
                statement
            ):
                return [statement]
            split = self.split_statement(statement.strip())
        return split

    def split_statement(self, text: str) -> list[str]:
        """
        Split the statement based on quotation marks position.

        Notes:
            If no quotation is found, uses tokenizer right away.
            If quotation is found, locates the first quoted triple and
            its surrounding terms and processes them separately, combining
            all in s/p/o list.

        Args:
            text (str): quoted triple to process

        Returns:
            list[str]: list of processed triple terms

        """
        start = text.find("<<")
        if start == -1:
            return self.generate_statement_tokens(text)
        pos = self.find_closing_quotation(text, start)

        if start == 0 and pos == len(text):
            triple = text[start + 2 : pos - 2].strip()
            split_triple = self.process_quoted(triple)
        else:
            split_triple = []
            if start > 0:
                text_before_quoted_triple = text[:start].strip()
                split_triple.extend(self.process_quoted(text_before_quoted_triple))
            split_triple.append(text[start:pos].strip())
            if pos < len(text):
                text_after_quoted_triple = text[pos:].strip()
                split_triple.extend(self.process_quoted(text_after_quoted_triple))
        return split_triple

    def generate_statement_tokens(self, statement: str) -> list[str]:
        """
        Split statement into separate tokens.

        Notes:
            Tokens are not validated to follow RDF spec.

        Args:
            statement (str): Triple/Quad to split.

        Returns:
            list[str]: tokens to further process, matching simple
                IRI, Literal, blank node, quoted triple formats.

        """
        return [
            m.group(0)
            for m in self._split_tokens.finditer(statement)
            if m.group(0).strip()
        ]

    def parse_statement(
        self, statement: str, statement_structure: type[Triple | Quad]
    ) -> Triple | Quad:
        """
        Create Triple/Quad from statement string.

        Args:
            statement (str): full statement string (triple/quad).
            statement_structure (type[Triple  |  Quad]): a data structure for statement.

        Returns:
            Triple | Quad: resulting triple/quad

        """
        terms = self.split_statement(statement)
        if len(terms) == TRIPLE_ARITY and statement_structure == Quad:
            terms.append("")  # append default graph identificator
        generic_terms = (
            self.process_term(term.strip(), slot)
            for slot, term in zip(statement_structure._fields, terms)
        )
        return statement_structure(*generic_terms)

    def parse_prefix(self, namespace: str) -> Prefix:
        """
        Create Prefix from namespace declaration string.

        Args:
            namespace (str): plain namespace declaration string.

        Raises:
            TypeError: raised if fail to match prefix and namespace IRI.

        Returns:
            Prefix: resulting Prefix

        """
        matched_namespace_declaration = self._prefix_re.match(namespace)
        if matched_namespace_declaration:
            matched_parts = matched_namespace_declaration.groups()
            return Prefix(matched_parts[0], IRI(matched_parts[1]))
        msg = "failed to parse namespace declaration"
        raise TypeError(msg)

    def parse_line(
        self, line: str, statement_structure: type[Triple | Quad]
    ) -> Triple | Quad | Prefix:
        """
        Parse one line from input into a respective statement or prefix.

        Args:
            line (str): plain line from source file.
            statement_structure (type[Triple  |  Quad]): a data structure for statement.

        Returns:
            Triple | Quad | Prefix: resulting parsed statement or namespace declaration.

        """
        if line.startswith("@prefix"):
            return self.parse_prefix(line)
        return self.parse_statement(line, statement_structure)

    def parse(self, input_filename: Path) -> None:
        """
        Parse input lines into statements and prefixes.

        Note:
            comments and blank lines in input file are ignored.

        """
        warnings.warn(
            (
                "This is a minimal parser for the NT/NQ format, "
                "not intended for use outside of conformance tests. "
                "Proceed with caution."
            ),
            category=UserWarning,
            stacklevel=2,
        )

        statement_structure = (
            Quad if str(input_filename).split(".")[-1] in ("nq") else Triple
        )
        with input_filename.open("r") as input_file:
            for line in input_file:
                line_trimmed = line[: line.rfind(".")] + line[line.rfind(".") + 1 :]
                line_trimmed = line_trimmed.strip()
                comment_index = line_trimmed.find("#")
                if comment_index == 0 or len(line_trimmed) == 0:
                    continue
                event = self.parse_line(line_trimmed, statement_structure)
                if isinstance(event, Prefix):
                    self._sink.bind(*event)
                else:
                    self._sink.add(event)



import pytest
from hypothesis import given
from hypothesis import strategies as st

from pyjelly.errors import JellyConformanceError
from pyjelly.options import LookupPreset


@given(st.integers(min_value=0, max_value=7))
def test_name_encoder_fails_with_size_lt_8(invalid_size: int) -> None:
    # max_name_table_size (9) - maximum size of the name lookup. This field is
    # REQUIRED and MUST be set to a value greater than or equal to 8. The size
    # of the lookup MUST NOT exceed the value of this field.
    with pytest.raises(JellyConformanceError, match="at least 8"):
        LookupPreset(
            max_names=invalid_size,
            max_prefixes=0,
            max_datatypes=0,
        )



from __future__ import annotations

import io
from unittest.mock import Mock

import pytest
from _pytest.monkeypatch import MonkeyPatch

import pyjelly.options as op
from pyjelly import jelly
from pyjelly.errors import JellyAssertionError, JellyConformanceError
from pyjelly.options import (
    MIN_NAME_LOOKUP_SIZE,
    MIN_VERSION,
    LookupPreset,
    StreamParameters,
    StreamTypes,
)
from pyjelly.parse.decode import options_from_frame
from pyjelly.parse.ioutils import get_options_and_frames


@pytest.mark.parametrize(
    ("physical_type", "logical_type"),
    VALID_STREAM_TYPE_COMBINATIONS := [
        (jelly.PHYSICAL_STREAM_TYPE_TRIPLES, jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES),
        (jelly.PHYSICAL_STREAM_TYPE_TRIPLES, jelly.LOGICAL_STREAM_TYPE_GRAPHS),
        (jelly.PHYSICAL_STREAM_TYPE_TRIPLES, jelly.LOGICAL_STREAM_TYPE_SUBJECT_GRAPHS),
        (jelly.PHYSICAL_STREAM_TYPE_TRIPLES, jelly.LOGICAL_STREAM_TYPE_UNSPECIFIED),
        (jelly.PHYSICAL_STREAM_TYPE_QUADS, jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS),
        (jelly.PHYSICAL_STREAM_TYPE_QUADS, jelly.LOGICAL_STREAM_TYPE_DATASETS),
        (jelly.PHYSICAL_STREAM_TYPE_QUADS, jelly.LOGICAL_STREAM_TYPE_NAMED_GRAPHS),
        (
            jelly.PHYSICAL_STREAM_TYPE_QUADS,
            jelly.LOGICAL_STREAM_TYPE_TIMESTAMPED_NAMED_GRAPHS,
        ),
        (jelly.PHYSICAL_STREAM_TYPE_QUADS, jelly.LOGICAL_STREAM_TYPE_UNSPECIFIED),
        (jelly.PHYSICAL_STREAM_TYPE_GRAPHS, jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS),
        (jelly.PHYSICAL_STREAM_TYPE_GRAPHS, jelly.LOGICAL_STREAM_TYPE_DATASETS),
        (jelly.PHYSICAL_STREAM_TYPE_GRAPHS, jelly.LOGICAL_STREAM_TYPE_NAMED_GRAPHS),
        (
            jelly.PHYSICAL_STREAM_TYPE_GRAPHS,
            jelly.LOGICAL_STREAM_TYPE_TIMESTAMPED_NAMED_GRAPHS,
        ),
        (jelly.PHYSICAL_STREAM_TYPE_GRAPHS, jelly.LOGICAL_STREAM_TYPE_UNSPECIFIED),
        (jelly.PHYSICAL_STREAM_TYPE_UNSPECIFIED, jelly.LOGICAL_STREAM_TYPE_UNSPECIFIED),
        (
            jelly.PHYSICAL_STREAM_TYPE_UNSPECIFIED,
            jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES,
        ),
        (jelly.PHYSICAL_STREAM_TYPE_UNSPECIFIED, jelly.LOGICAL_STREAM_TYPE_GRAPHS),
        (
            jelly.PHYSICAL_STREAM_TYPE_UNSPECIFIED,
            jelly.LOGICAL_STREAM_TYPE_SUBJECT_GRAPHS,
        ),
        (jelly.PHYSICAL_STREAM_TYPE_UNSPECIFIED, jelly.LOGICAL_STREAM_TYPE_UNSPECIFIED),
        (jelly.PHYSICAL_STREAM_TYPE_UNSPECIFIED, jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS),
        (jelly.PHYSICAL_STREAM_TYPE_UNSPECIFIED, jelly.LOGICAL_STREAM_TYPE_DATASETS),
        (
            jelly.PHYSICAL_STREAM_TYPE_UNSPECIFIED,
            jelly.LOGICAL_STREAM_TYPE_NAMED_GRAPHS,
        ),
        (
            jelly.PHYSICAL_STREAM_TYPE_UNSPECIFIED,
            jelly.LOGICAL_STREAM_TYPE_TIMESTAMPED_NAMED_GRAPHS,
        ),
    ],
)
def test_stream_types_ok(
    physical_type: jelly.PhysicalStreamType,
    logical_type: jelly.LogicalStreamType,
) -> None:
    StreamTypes(physical_type=physical_type, logical_type=logical_type)


@pytest.mark.parametrize(
    ("physical_type", "logical_type"),
    [
        (jelly.PHYSICAL_STREAM_TYPE_TRIPLES, jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS),
        (jelly.PHYSICAL_STREAM_TYPE_TRIPLES, jelly.LOGICAL_STREAM_TYPE_DATASETS),
        (jelly.PHYSICAL_STREAM_TYPE_TRIPLES, jelly.LOGICAL_STREAM_TYPE_NAMED_GRAPHS),
        (
            jelly.PHYSICAL_STREAM_TYPE_TRIPLES,
            jelly.LOGICAL_STREAM_TYPE_TIMESTAMPED_NAMED_GRAPHS,
        ),
        (jelly.PHYSICAL_STREAM_TYPE_QUADS, jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES),
        (jelly.PHYSICAL_STREAM_TYPE_QUADS, jelly.LOGICAL_STREAM_TYPE_GRAPHS),
        (jelly.PHYSICAL_STREAM_TYPE_QUADS, jelly.LOGICAL_STREAM_TYPE_SUBJECT_GRAPHS),
        (jelly.PHYSICAL_STREAM_TYPE_GRAPHS, jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES),
        (jelly.PHYSICAL_STREAM_TYPE_GRAPHS, jelly.LOGICAL_STREAM_TYPE_GRAPHS),
        (jelly.PHYSICAL_STREAM_TYPE_GRAPHS, jelly.LOGICAL_STREAM_TYPE_SUBJECT_GRAPHS),
    ],
)
def test_stream_types_error(
    physical_type: jelly.PhysicalStreamType,
    logical_type: jelly.LogicalStreamType,
) -> None:
    with pytest.raises(JellyAssertionError):
        StreamTypes(physical_type=physical_type, logical_type=logical_type)


@pytest.mark.parametrize(
    ("generalized_statements", "rdf_star"), [(0, 0), (0, 1), (1, 0), (1, 1)]
)
def test_optional_fields(generalized_statements: int, rdf_star: int) -> None:
    mock_frame = Mock()
    mock_row = Mock()
    mock_options = Mock()

    defaults = {
        "physical_type": jelly.PHYSICAL_STREAM_TYPE_TRIPLES,
        "logical_type": jelly.PHYSICAL_STREAM_TYPE_UNSPECIFIED,
        "max_name_table_size": 1000,
        "max_prefix_table_size": 100,
        "max_datatype_table_size": 100,
        "stream_name": "",
        "generalized_statements": bool(generalized_statements),
        "rdf_star": bool(rdf_star),
        "version": 1,
    }

    for key, value in defaults.items():
        setattr(mock_options, key, value)

    mock_row.options = mock_options
    mock_frame.rows = [mock_row]

    result = options_from_frame(mock_frame, delimited=True)

    assert result.params.generalized_statements == defaults["generalized_statements"]
    assert result.params.rdf_star == defaults["rdf_star"]


@pytest.mark.parametrize(
    ("generalized_statements", "rdf_star"), [(0, 0), (0, 1), (1, 0), (1, 1)]
)
def test_stream_parameters(generalized_statements: int, rdf_star: int) -> None:
    mock_options = Mock()

    defaults = {
        "physical_type": jelly.PHYSICAL_STREAM_TYPE_TRIPLES,
        "logical_type": jelly.PHYSICAL_STREAM_TYPE_UNSPECIFIED,
        "max_name_table_size": 1000,
        "max_prefix_table_size": 100,
        "max_datatype_table_size": 100,
        "stream_name": "",
        "generalized_statements": bool(generalized_statements),
        "rdf_star": bool(rdf_star),
        "version": 1,
    }

    for key, value in defaults.items():
        setattr(mock_options, key, value)

    params = StreamParameters(
        stream_name=mock_options.stream_name,
        generalized_statements=mock_options.generalized_statements,
        rdf_star=mock_options.rdf_star,
        version=mock_options.version,
        delimited=True,
    )
    assert params.generalized_statements == mock_options.generalized_statements
    assert params.rdf_star == mock_options.rdf_star


def test_lookup_preset_validation() -> None:
    with pytest.raises(JellyConformanceError):
        LookupPreset(max_names=MIN_NAME_LOOKUP_SIZE - 1)
    p = LookupPreset.small()
    assert (p.max_names, p.max_prefixes, p.max_datatypes) == (128, 32, 32)


def test_stream_parameters_version() -> None:
    s1 = StreamParameters(namespace_declarations=False)
    assert s1.version == MIN_VERSION
    s2 = StreamParameters(namespace_declarations=True)
    assert s2.version == 2


def test_stream_options_invalid_version(monkeypatch: MonkeyPatch) -> None:
    monkeypatch.setattr(op, "MIN_VERSION", 10)
    monkeypatch.setattr(op, "MAX_VERSION", 5)

    with pytest.raises(JellyConformanceError) as excinfo:
        StreamParameters(namespace_declarations=True)

    msg = str(excinfo.value)
    assert "Version must be between 10 and 5" in msg


def test_get_options_and_frames_delimited_only_empty() -> None:
    with pytest.raises(JellyConformanceError) as exc:
        get_options_and_frames(io.BytesIO(b"\x00\x00\x00"))
    assert "No non-empty frames found" in str(exc.value)


def test_get_options_and_frames_non_delimited_empty() -> None:
    with pytest.raises(JellyConformanceError) as exc:
        get_options_and_frames(io.BytesIO(b""))
    assert "only contains an empty frame" in str(exc.value)


def test_get_options_and_frames_non_delimited_success_only_return() -> None:
    opts_msg = jelly.RdfStreamOptions(
        stream_name="test",
        physical_type=jelly.PHYSICAL_STREAM_TYPE_TRIPLES,
        generalized_statements=False,
        rdf_star=False,
        max_name_table_size=8,
        max_prefix_table_size=8,
        max_datatype_table_size=8,
        logical_type=jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES,
        version=1,
    )
    frame = jelly.RdfStreamFrame(rows=[jelly.RdfStreamRow(options=opts_msg)])
    data = frame.SerializeToString(deterministic=True)

    options, frames_iter = get_options_and_frames(io.BytesIO(data))

    assert options.stream_types.physical_type == jelly.PHYSICAL_STREAM_TYPE_TRIPLES
    assert options.stream_types.logical_type == jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES

    frames = list(frames_iter)
    assert len(frames) == 1
    assert isinstance(frames[0], jelly.RdfStreamFrame)
    assert frames[0].rows[0].options == opts_msg



import rdflib
import rdflib.util

from pyjelly import options


def register_extension_to_rdflib(extension: str = ".jelly") -> None:
    """
    Make [rdflib.util.guess_format][] discover Jelly format.

    >>> rdflib.util.guess_format("foo.jelly")
    >>> register_extension_to_rdflib()
    >>> rdflib.util.guess_format("foo.jelly")
    'jelly'
    """
    rdflib.util.SUFFIX_FORMAT_MAP[extension.removeprefix(".")] = "jelly"


def _side_effects() -> None:
    register_extension_to_rdflib()


if options.INTEGRATION_SIDE_EFFECTS:
    _side_effects()



import gzip
import urllib.request

from pyjelly.integrations.rdflib.parse import parse_jelly_flat, Triple
from rdflib import URIRef

# Dataset: OpenStreetMap data for Denmark (first 10k objects)
# Documentation: https://w3id.org/riverbench/datasets/osm2rdf-denmark/dev
url = (
    "https://w3id.org/riverbench/datasets/osm2rdf-denmark/dev/files/jelly_10K.jelly.gz"
)

# We are looking for city names in the dataset
predicate_to_look_for = URIRef("https://www.openstreetmap.org/wiki/Key:addr:city")
city_names = set()

with (
    urllib.request.urlopen(url) as response,
    gzip.open(response) as jelly_stream,
):
    for event in parse_jelly_flat(jelly_stream):
        if isinstance(event, Triple):  # we are only interested in triples
            if event.p == predicate_to_look_for:
                city_names.add(event.o)

print(f"Found {len(city_names)} unique city names in the dataset.")
print("10 random city names:")
for city in list(city_names)[:10]:
    print(f"- {city}")



from __future__ import annotations

import shlex
import shutil
import subprocess
from enum import Enum
from functools import partial
from pathlib import Path

import pytest

# TODO(Piotr): Remove this list once the failing test cases are fixed.
# See https://github.com/Jelly-RDF/pyjelly/issues/145
failing_test_cases = [
    "to_jelly_triples_rdf_1_1_pos_014",
]

JELLY_CLI = shutil.which("jelly-cli")

needs_jelly_cli = pytest.mark.skipif(
    not JELLY_CLI,
    reason="jelly-cli not found in PATH",
)


class JellyCLIError(Exception):
    """Exception raised when jelly-cli command fails."""


def jelly_cli(*args: object, hint: str | None = None) -> bytes:
    assert JELLY_CLI
    shell_args = [JELLY_CLI, *map(str, args)]
    try:
        return subprocess.check_output(shell_args, stderr=subprocess.STDOUT)  # noqa: S603 internal use
    except subprocess.CalledProcessError as error:
        command = shlex.join(shell_args)
        note = f"Command: {command}"
        if hint:
            note += f"\nHint: {hint}"
        raise JellyCLIError(error.output.decode() + "\n" + note) from None


jelly_validate = partial(jelly_cli, "rdf", "validate")


def id_from_path(path: Path) -> str:
    return f"{path.parent.parent.name}_{path.parent.name}_{path.name}"


class PhysicalTypeTestCasesDir(str, Enum):
    TRIPLES = "triples_rdf_1_1"
    QUADS = "quads_rdf_1_1"
    GRAPHS = "graphs_rdf_1_1"

    def __str__(self) -> str:
        return self.value


class GeneralizedTestCasesDir(str, Enum):
    TRIPLES = "triples_rdf_1_1_generalized"
    QUADS = "quads_rdf_1_1_generalized"
    GRAPHS = "graphs_rdf_1_1_generalized"

    def __str__(self) -> str:
        return self.value


class RDFStarGeneralizedTestCasesDir(str, Enum):
    TRIPLES = "triples_rdf_star_generalized"
    QUADS = "quads_rdf_star_generalized"
    GRAPHS = "graphs_rdf_star_generalized"

    def __str__(self) -> str:
        return self.value


class RDFStarTestCasesDir(str, Enum):
    TRIPLES = "triples_rdf_star"
    QUADS = "quads_rdf_star"
    GRAPHS = "graphs_rdf_star"

    def __str__(self) -> str:
        return self.value


def walk_directories(
    *dirs: str | Path,
    glob: str | None = None,
) -> pytest.MarkDecorator:
    paths: list[Path] = []

    for directory in map(Path, dirs):
        if not directory.is_dir():
            # a warning here, albeit potentially helpful, is too noisy in practice
            continue
        paths.extend(directory.glob(glob or "*"))

    paths = [path for path in paths if id_from_path(path) not in failing_test_cases]
    if not paths:
        return pytest.mark.parametrize("path", [None], ids=["no-tests-found"])

    return pytest.mark.parametrize("path", paths, ids=id_from_path)



from pathlib import Path

TESTS_DIR = Path(__file__).parent
CONFORMANCE_TESTS_DIR = TESTS_DIR / "conformance_tests"
E2E_TEST_CASES_DIR = TESTS_DIR / "e2e_test_cases"

TEST_OUTPUTS_DIR = TESTS_DIR / "out"

RDF_TESTS_DIR = CONFORMANCE_TESTS_DIR / "rdf"
RDF_FROM_JELLY_TESTS_DIR = RDF_TESTS_DIR / "from_jelly"
RDF_TO_JELLY_TESTS_DIR = RDF_TESTS_DIR / "to_jelly"



from __future__ import annotations

import sys

import pytest
import rdflib

from pyjelly import options

# No need for protoletariat ;)
sys.path.append("pyjelly/_pb2/")

options.INTEGRATION_SIDE_EFFECTS = False
rdflib.NORMALIZE_LITERALS = False


@pytest.fixture(autouse=True)
def make_output_dir() -> None:
    """Create the output directory if it doesn't exist."""
    from tests.meta import TEST_OUTPUTS_DIR

    TEST_OUTPUTS_DIR.mkdir(exist_ok=True)



from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from collections.abc import Iterable as _Iterable, Mapping as _Mapping
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class PhysicalStreamType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    PHYSICAL_STREAM_TYPE_UNSPECIFIED: _ClassVar[PhysicalStreamType]
    PHYSICAL_STREAM_TYPE_TRIPLES: _ClassVar[PhysicalStreamType]
    PHYSICAL_STREAM_TYPE_QUADS: _ClassVar[PhysicalStreamType]
    PHYSICAL_STREAM_TYPE_GRAPHS: _ClassVar[PhysicalStreamType]

class LogicalStreamType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    LOGICAL_STREAM_TYPE_UNSPECIFIED: _ClassVar[LogicalStreamType]
    LOGICAL_STREAM_TYPE_FLAT_TRIPLES: _ClassVar[LogicalStreamType]
    LOGICAL_STREAM_TYPE_FLAT_QUADS: _ClassVar[LogicalStreamType]
    LOGICAL_STREAM_TYPE_GRAPHS: _ClassVar[LogicalStreamType]
    LOGICAL_STREAM_TYPE_DATASETS: _ClassVar[LogicalStreamType]
    LOGICAL_STREAM_TYPE_SUBJECT_GRAPHS: _ClassVar[LogicalStreamType]
    LOGICAL_STREAM_TYPE_NAMED_GRAPHS: _ClassVar[LogicalStreamType]
    LOGICAL_STREAM_TYPE_TIMESTAMPED_NAMED_GRAPHS: _ClassVar[LogicalStreamType]
PHYSICAL_STREAM_TYPE_UNSPECIFIED: PhysicalStreamType
PHYSICAL_STREAM_TYPE_TRIPLES: PhysicalStreamType
PHYSICAL_STREAM_TYPE_QUADS: PhysicalStreamType
PHYSICAL_STREAM_TYPE_GRAPHS: PhysicalStreamType
LOGICAL_STREAM_TYPE_UNSPECIFIED: LogicalStreamType
LOGICAL_STREAM_TYPE_FLAT_TRIPLES: LogicalStreamType
LOGICAL_STREAM_TYPE_FLAT_QUADS: LogicalStreamType
LOGICAL_STREAM_TYPE_GRAPHS: LogicalStreamType
LOGICAL_STREAM_TYPE_DATASETS: LogicalStreamType
LOGICAL_STREAM_TYPE_SUBJECT_GRAPHS: LogicalStreamType
LOGICAL_STREAM_TYPE_NAMED_GRAPHS: LogicalStreamType
LOGICAL_STREAM_TYPE_TIMESTAMPED_NAMED_GRAPHS: LogicalStreamType

class RdfIri(_message.Message):
    __slots__ = ("prefix_id", "name_id")
    PREFIX_ID_FIELD_NUMBER: _ClassVar[int]
    NAME_ID_FIELD_NUMBER: _ClassVar[int]
    prefix_id: int
    name_id: int
    def __init__(self, prefix_id: _Optional[int] = ..., name_id: _Optional[int] = ...) -> None: ...

class RdfLiteral(_message.Message):
    __slots__ = ("lex", "langtag", "datatype")
    LEX_FIELD_NUMBER: _ClassVar[int]
    LANGTAG_FIELD_NUMBER: _ClassVar[int]
    DATATYPE_FIELD_NUMBER: _ClassVar[int]
    lex: str
    langtag: str
    datatype: int
    def __init__(self, lex: _Optional[str] = ..., langtag: _Optional[str] = ..., datatype: _Optional[int] = ...) -> None: ...

class RdfDefaultGraph(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class RdfTriple(_message.Message):
    __slots__ = ("s_iri", "s_bnode", "s_literal", "s_triple_term", "p_iri", "p_bnode", "p_literal", "p_triple_term", "o_iri", "o_bnode", "o_literal", "o_triple_term")
    S_IRI_FIELD_NUMBER: _ClassVar[int]
    S_BNODE_FIELD_NUMBER: _ClassVar[int]
    S_LITERAL_FIELD_NUMBER: _ClassVar[int]
    S_TRIPLE_TERM_FIELD_NUMBER: _ClassVar[int]
    P_IRI_FIELD_NUMBER: _ClassVar[int]
    P_BNODE_FIELD_NUMBER: _ClassVar[int]
    P_LITERAL_FIELD_NUMBER: _ClassVar[int]
    P_TRIPLE_TERM_FIELD_NUMBER: _ClassVar[int]
    O_IRI_FIELD_NUMBER: _ClassVar[int]
    O_BNODE_FIELD_NUMBER: _ClassVar[int]
    O_LITERAL_FIELD_NUMBER: _ClassVar[int]
    O_TRIPLE_TERM_FIELD_NUMBER: _ClassVar[int]
    s_iri: RdfIri
    s_bnode: str
    s_literal: RdfLiteral
    s_triple_term: RdfTriple
    p_iri: RdfIri
    p_bnode: str
    p_literal: RdfLiteral
    p_triple_term: RdfTriple
    o_iri: RdfIri
    o_bnode: str
    o_literal: RdfLiteral
    o_triple_term: RdfTriple
    def __init__(self, s_iri: _Optional[_Union[RdfIri, _Mapping]] = ..., s_bnode: _Optional[str] = ..., s_literal: _Optional[_Union[RdfLiteral, _Mapping]] = ..., s_triple_term: _Optional[_Union[RdfTriple, _Mapping]] = ..., p_iri: _Optional[_Union[RdfIri, _Mapping]] = ..., p_bnode: _Optional[str] = ..., p_literal: _Optional[_Union[RdfLiteral, _Mapping]] = ..., p_triple_term: _Optional[_Union[RdfTriple, _Mapping]] = ..., o_iri: _Optional[_Union[RdfIri, _Mapping]] = ..., o_bnode: _Optional[str] = ..., o_literal: _Optional[_Union[RdfLiteral, _Mapping]] = ..., o_triple_term: _Optional[_Union[RdfTriple, _Mapping]] = ...) -> None: ...

class RdfQuad(_message.Message):
    __slots__ = ("s_iri", "s_bnode", "s_literal", "s_triple_term", "p_iri", "p_bnode", "p_literal", "p_triple_term", "o_iri", "o_bnode", "o_literal", "o_triple_term", "g_iri", "g_bnode", "g_default_graph", "g_literal")
    S_IRI_FIELD_NUMBER: _ClassVar[int]
    S_BNODE_FIELD_NUMBER: _ClassVar[int]
    S_LITERAL_FIELD_NUMBER: _ClassVar[int]
    S_TRIPLE_TERM_FIELD_NUMBER: _ClassVar[int]
    P_IRI_FIELD_NUMBER: _ClassVar[int]
    P_BNODE_FIELD_NUMBER: _ClassVar[int]
    P_LITERAL_FIELD_NUMBER: _ClassVar[int]
    P_TRIPLE_TERM_FIELD_NUMBER: _ClassVar[int]
    O_IRI_FIELD_NUMBER: _ClassVar[int]
    O_BNODE_FIELD_NUMBER: _ClassVar[int]
    O_LITERAL_FIELD_NUMBER: _ClassVar[int]
    O_TRIPLE_TERM_FIELD_NUMBER: _ClassVar[int]
    G_IRI_FIELD_NUMBER: _ClassVar[int]
    G_BNODE_FIELD_NUMBER: _ClassVar[int]
    G_DEFAULT_GRAPH_FIELD_NUMBER: _ClassVar[int]
    G_LITERAL_FIELD_NUMBER: _ClassVar[int]
    s_iri: RdfIri
    s_bnode: str
    s_literal: RdfLiteral
    s_triple_term: RdfTriple
    p_iri: RdfIri
    p_bnode: str
    p_literal: RdfLiteral
    p_triple_term: RdfTriple
    o_iri: RdfIri
    o_bnode: str
    o_literal: RdfLiteral
    o_triple_term: RdfTriple
    g_iri: RdfIri
    g_bnode: str
    g_default_graph: RdfDefaultGraph
    g_literal: RdfLiteral
    def __init__(self, s_iri: _Optional[_Union[RdfIri, _Mapping]] = ..., s_bnode: _Optional[str] = ..., s_literal: _Optional[_Union[RdfLiteral, _Mapping]] = ..., s_triple_term: _Optional[_Union[RdfTriple, _Mapping]] = ..., p_iri: _Optional[_Union[RdfIri, _Mapping]] = ..., p_bnode: _Optional[str] = ..., p_literal: _Optional[_Union[RdfLiteral, _Mapping]] = ..., p_triple_term: _Optional[_Union[RdfTriple, _Mapping]] = ..., o_iri: _Optional[_Union[RdfIri, _Mapping]] = ..., o_bnode: _Optional[str] = ..., o_literal: _Optional[_Union[RdfLiteral, _Mapping]] = ..., o_triple_term: _Optional[_Union[RdfTriple, _Mapping]] = ..., g_iri: _Optional[_Union[RdfIri, _Mapping]] = ..., g_bnode: _Optional[str] = ..., g_default_graph: _Optional[_Union[RdfDefaultGraph, _Mapping]] = ..., g_literal: _Optional[_Union[RdfLiteral, _Mapping]] = ...) -> None: ...

class RdfGraphStart(_message.Message):
    __slots__ = ("g_iri", "g_bnode", "g_default_graph", "g_literal")
    G_IRI_FIELD_NUMBER: _ClassVar[int]
    G_BNODE_FIELD_NUMBER: _ClassVar[int]
    G_DEFAULT_GRAPH_FIELD_NUMBER: _ClassVar[int]
    G_LITERAL_FIELD_NUMBER: _ClassVar[int]
    g_iri: RdfIri
    g_bnode: str
    g_default_graph: RdfDefaultGraph
    g_literal: RdfLiteral
    def __init__(self, g_iri: _Optional[_Union[RdfIri, _Mapping]] = ..., g_bnode: _Optional[str] = ..., g_default_graph: _Optional[_Union[RdfDefaultGraph, _Mapping]] = ..., g_literal: _Optional[_Union[RdfLiteral, _Mapping]] = ...) -> None: ...

class RdfGraphEnd(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class RdfNamespaceDeclaration(_message.Message):
    __slots__ = ("name", "value")
    NAME_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    name: str
    value: RdfIri
    def __init__(self, name: _Optional[str] = ..., value: _Optional[_Union[RdfIri, _Mapping]] = ...) -> None: ...

class RdfNameEntry(_message.Message):
    __slots__ = ("id", "value")
    ID_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    id: int
    value: str
    def __init__(self, id: _Optional[int] = ..., value: _Optional[str] = ...) -> None: ...

class RdfPrefixEntry(_message.Message):
    __slots__ = ("id", "value")
    ID_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    id: int
    value: str
    def __init__(self, id: _Optional[int] = ..., value: _Optional[str] = ...) -> None: ...

class RdfDatatypeEntry(_message.Message):
    __slots__ = ("id", "value")
    ID_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    id: int
    value: str
    def __init__(self, id: _Optional[int] = ..., value: _Optional[str] = ...) -> None: ...

class RdfStreamOptions(_message.Message):
    __slots__ = ("stream_name", "physical_type", "generalized_statements", "rdf_star", "max_name_table_size", "max_prefix_table_size", "max_datatype_table_size", "logical_type", "version")
    STREAM_NAME_FIELD_NUMBER: _ClassVar[int]
    PHYSICAL_TYPE_FIELD_NUMBER: _ClassVar[int]
    GENERALIZED_STATEMENTS_FIELD_NUMBER: _ClassVar[int]
    RDF_STAR_FIELD_NUMBER: _ClassVar[int]
    MAX_NAME_TABLE_SIZE_FIELD_NUMBER: _ClassVar[int]
    MAX_PREFIX_TABLE_SIZE_FIELD_NUMBER: _ClassVar[int]
    MAX_DATATYPE_TABLE_SIZE_FIELD_NUMBER: _ClassVar[int]
    LOGICAL_TYPE_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    stream_name: str
    physical_type: PhysicalStreamType
    generalized_statements: bool
    rdf_star: bool
    max_name_table_size: int
    max_prefix_table_size: int
    max_datatype_table_size: int
    logical_type: LogicalStreamType
    version: int
    def __init__(self, stream_name: _Optional[str] = ..., physical_type: _Optional[_Union[PhysicalStreamType, str]] = ..., generalized_statements: bool = ..., rdf_star: bool = ..., max_name_table_size: _Optional[int] = ..., max_prefix_table_size: _Optional[int] = ..., max_datatype_table_size: _Optional[int] = ..., logical_type: _Optional[_Union[LogicalStreamType, str]] = ..., version: _Optional[int] = ...) -> None: ...

class RdfStreamRow(_message.Message):
    __slots__ = ("options", "triple", "quad", "graph_start", "graph_end", "namespace", "name", "prefix", "datatype")
    OPTIONS_FIELD_NUMBER: _ClassVar[int]
    TRIPLE_FIELD_NUMBER: _ClassVar[int]
    QUAD_FIELD_NUMBER: _ClassVar[int]
    GRAPH_START_FIELD_NUMBER: _ClassVar[int]
    GRAPH_END_FIELD_NUMBER: _ClassVar[int]
    NAMESPACE_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    PREFIX_FIELD_NUMBER: _ClassVar[int]
    DATATYPE_FIELD_NUMBER: _ClassVar[int]
    options: RdfStreamOptions
    triple: RdfTriple
    quad: RdfQuad
    graph_start: RdfGraphStart
    graph_end: RdfGraphEnd
    namespace: RdfNamespaceDeclaration
    name: RdfNameEntry
    prefix: RdfPrefixEntry
    datatype: RdfDatatypeEntry
    def __init__(self, options: _Optional[_Union[RdfStreamOptions, _Mapping]] = ..., triple: _Optional[_Union[RdfTriple, _Mapping]] = ..., quad: _Optional[_Union[RdfQuad, _Mapping]] = ..., graph_start: _Optional[_Union[RdfGraphStart, _Mapping]] = ..., graph_end: _Optional[_Union[RdfGraphEnd, _Mapping]] = ..., namespace: _Optional[_Union[RdfNamespaceDeclaration, _Mapping]] = ..., name: _Optional[_Union[RdfNameEntry, _Mapping]] = ..., prefix: _Optional[_Union[RdfPrefixEntry, _Mapping]] = ..., datatype: _Optional[_Union[RdfDatatypeEntry, _Mapping]] = ...) -> None: ...

class RdfStreamFrame(_message.Message):
    __slots__ = ("rows", "metadata")
    class MetadataEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: bytes
        def __init__(self, key: _Optional[str] = ..., value: _Optional[bytes] = ...) -> None: ...
    ROWS_FIELD_NUMBER: _ClassVar[int]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    rows: _containers.RepeatedCompositeFieldContainer[RdfStreamRow]
    metadata: _containers.ScalarMap[str, bytes]
    def __init__(self, rows: _Optional[_Iterable[_Union[RdfStreamRow, _Mapping]]] = ..., metadata: _Optional[_Mapping[str, bytes]] = ...) -> None: ...



import urllib.request, gzip, tempfile, shutil, os
from rdflib import Graph
from rdflib.extras.external_graph_libs import rdflib_to_networkx_graph
import networkx as nx
import matplotlib.pyplot as plt
from rdflib import Graph, URIRef, BNode, Literal

# Filter predicates for people
TARGET_PREDICATES = {
    URIRef("http://www.politiquices.pt/ent2_str"),
    URIRef("http://www.politiquices.pt/ent1_str"),
}

# URL to the dataset
url = "https://w3id.org/riverbench/datasets/politiquices/1.0.3/files/jelly_10K.jelly.gz"

# Load example jelly file
with urllib.request.urlopen(url) as r:
    fd, example_file = tempfile.mkstemp(suffix=".jelly")
    os.close(fd)
    with gzip.GzipFile(fileobj=r) as g, open(example_file, "wb") as out:
        shutil.copyfileobj(g, out)

# Parse RDF from the Jelly format
rdf_g = Graph()
rdf_g.parse(example_file, format="jelly")
print(f"Loaded graph with {len(rdf_g)} instances.")

# First step for filtering the graph
rdf_g_filtered = Graph()
for s, p, o in rdf_g.triples((None, None, None)):
    if p in TARGET_PREDICATES:
        rdf_g_filtered.add((s, p, o))

# Just filtering the graph for clarity of visualization, can be omitted
related = URIRef("http://www.politiquices.pt/related")
for b in list(rdf_g_filtered.subjects()):
    if isinstance(b, BNode):
        ent1 = ent2 = None
        for p, o in rdf_g_filtered.predicate_objects(b):
            if p == URIRef("http://www.politiquices.pt/ent1_str") and isinstance(
                o, Literal
            ):
                ent1 = o
            elif p == URIRef("http://www.politiquices.pt/ent2_str") and isinstance(
                o, Literal
            ):
                ent2 = o
        if ent1 and ent2:
            rdf_g_filtered.add((ent1, related, ent2))
            rdf_g_filtered.add((ent2, related, ent1))
        for triple in list(rdf_g_filtered.triples((b, None, None))):
            rdf_g_filtered.remove(triple)

# Convert to a NetworkX graph
nx_g = rdflib_to_networkx_graph(rdf_g_filtered)

# Example calculation, get the number of connected components in a graph
num_components = nx.number_connected_components(nx_g)
print(f"Connected components: {num_components}")

# Example calculation, get top 5 objects with highest degrees, simple in NetworkX
top5 = sorted(nx_g.degree, key=lambda x: x[1], reverse=True)[:5]
print("Top 5 nodes sorted by degree:")
for node, deg in top5:
    print(f"{node}: {deg}")

# Helper function
norm = (
    lambda n: str(n.value).strip().lower()
    if isinstance(n, Literal)
    else str(n).strip().lower()
)

# Example calculation, shortest path between two nodes (provided at least two nodes)
source = next(n for n in nx_g if norm(n) == "socrates")
target = next(n for n in nx_g if norm(n) == "obama")
path = nx.shortest_path(nx_g, source=source, target=target)
print(f"Shortest path from {source} to {target}: {' -> '.join(path)}")

# Take first 10 nodes
nodes = list(nx_g)[:10]
subg = nx_g.subgraph(nodes)

# Draw and display the graph
pos_sub = nx.spring_layout(subg, k=5, iterations=200, scale=3, seed=24)
plt.figure(figsize=(10, 10))

# Introduce your own settings for display
nx.draw_networkx(subg, pos_sub, font_size=14, node_size=220, linewidths=0.7)
plt.axis("off")
plt.show()

print("All done.")



from __future__ import annotations

from pathlib import Path

from pyjelly.integrations.generic.generic_sink import (
    GenericStatementSink,
    Node,
    Triple,
)


class GenericSinkSerializer:
    def __init__(self, sink: GenericStatementSink) -> None:
        self._sink = sink

    def _serialize_node(self, node: Node) -> str:
        """
        Serialize node to its string representation.

        Args:
            node (Node): Node to convert - RDF term, str, or Triple.

        Returns:
            str: string representation of Node.

        """
        if isinstance(node, Triple):
            quoted_triple = [self._serialize_node(t) for t in node]
            return "<< " + " ".join(quoted_triple) + " >>"
        return str(node)

    def serialize(self, output_filename: Path, encoding: str = "utf-8") -> None:
        """
        Serialize sink's store content to a simple N-triples/N-quads format.

        Args:
            output_filename (Path): path to the output file.
            encoding (str): encoding of output. Defaults to utf-8.

        """
        with output_filename.open("w", encoding=encoding) as output_file:
            for statement in self._sink._store:
                output_file.write(
                    " ".join(self._serialize_node(t) for t in statement) + " .\n"
                )



# ruff: noqa: I001
from __future__ import annotations
from typing import cast
from collections.abc import Generator
from functools import singledispatch
from typing import Any, IO
from itertools import chain
from pyjelly.integrations.generic.generic_sink import (
    GenericStatementSink,
    Quad,
    Triple,
    DEFAULT_GRAPH_IDENTIFIER,
    IRI,
    BlankNode,
    Literal,
    Node,
)

from pyjelly import jelly
from pyjelly.serialize.encode import RowsAndTerm, Slot, TermEncoder
from pyjelly.serialize.ioutils import write_delimited
from pyjelly.serialize.streams import (
    GraphStream,
    QuadStream,
    SerializerOptions,
    Stream,
    TripleStream,
)  # ruff: enable

QUAD_ARITY = 4


class GenericSinkTermEncoder(TermEncoder):
    def encode_any(self, term: object, slot: Slot) -> RowsAndTerm:
        """
        Encode term based on its GenericSink object.

        Args:
            term (object): term to encode
            slot (Slot): its place in statement.

        Returns:
            RowsAndTerm: encoded extra rows and a jelly term to encode

        """
        if slot is Slot.graph and term == DEFAULT_GRAPH_IDENTIFIER:
            return self.encode_default_graph()

        if isinstance(term, IRI):
            return self.encode_iri(term._iri)

        if isinstance(term, Literal):
            return self.encode_literal(
                lex=term._lex,
                language=term._langtag,
                datatype=term._datatype,
            )

        if isinstance(term, BlankNode):
            return self.encode_bnode(term._identifier)

        if isinstance(term, Triple):
            return self.encode_quoted_triple(term)

        return super().encode_any(term, slot)  # error if not handled


def namespace_declarations(store: GenericStatementSink, stream: Stream) -> None:
    for prefix, namespace in store.namespaces:
        stream.namespace_declaration(name=prefix, iri=str(namespace))


@singledispatch
def stream_frames(
    stream: Stream,
    data: GenericStatementSink | Generator[Quad | Triple],  # noqa: ARG001
) -> Generator[jelly.RdfStreamFrame]:
    msg = f"invalid stream implementation {stream}"
    raise TypeError(msg)


@stream_frames.register(TripleStream)
def triples_stream_frames(
    stream: TripleStream,
    data: GenericStatementSink | Generator[Triple],
) -> Generator[jelly.RdfStreamFrame]:
    """
    Serialize a GenericStatementSink into frames using physical type triples stream.

    Args:
        stream (TripleStream): stream that specifies triples processing
        data (GenericStatementSink | Generator[Triple]):
            GenericStatementSink/Statements to serialize.

    Yields:
        Generator[jelly.RdfStreamFrame]: jelly frames.

    """
    stream.enroll()
    if (
        isinstance(data, GenericStatementSink)
        and stream.options.params.namespace_declarations
    ):
        namespace_declarations(data, stream)

    graphs = (data,)
    for graph in graphs:
        for terms in graph:
            if frame := stream.triple(terms):
                yield frame
        if frame := stream.flow.frame_from_graph():
            yield frame
    if stream.stream_types.flat and (frame := stream.flow.to_stream_frame()):
        yield frame


@stream_frames.register(QuadStream)
def quads_stream_frames(
    stream: QuadStream,
    data: GenericStatementSink | Generator[Quad],
) -> Generator[jelly.RdfStreamFrame]:
    """
    Serialize a GenericStatementSink into jelly frames using physical type quads stream.

    Args:
        stream (QuadStream): stream that specifies quads processing
        data (GenericStatementSink | Generator[Quad]): Dataset to serialize.

    Yields:
        Generator[jelly.RdfStreamFrame]: jelly frames

    """
    stream.enroll()
    if stream.options.params.namespace_declarations:
        namespace_declarations(data, stream)  # type: ignore[arg-type]

    iterator: Generator[Quad]
    if isinstance(data, GenericStatementSink):
        iterator = cast(Generator[Quad], data.store)
    else:
        iterator = data

    for terms in iterator:
        if frame := stream.quad(terms):
            yield frame
    if frame := stream.flow.frame_from_dataset():
        yield frame
    if stream.stream_types.flat and (frame := stream.flow.to_stream_frame()):
        yield frame


@stream_frames.register(GraphStream)
def graphs_stream_frames(
    stream: GraphStream,
    data: GenericStatementSink | Generator[Quad],
) -> Generator[jelly.RdfStreamFrame]:
    """
    Serialize a GenericStatementSink into jelly frames as a stream of graphs.

    Notes:
        If flow of DatasetsFrameFlow type, the whole dataset
        will be encoded into one frame.
        Graphs are generated from the GenericStatementSink by
        iterating over statements and yielding one new GenericStatementSink
        per a sequence of quads with the same g term.

    Args:
        stream (GraphStream): stream that specifies graphs processing
        data (GenericStatementSink | Generator[Quad]): Dataset to serialize.

    Yields:
        Generator[jelly.RdfStreamFrame]: jelly frames

    """
    stream.enroll()
    if stream.options.params.namespace_declarations:
        namespace_declarations(data, stream)  # type: ignore[arg-type]

    statements: Generator[Quad]
    if isinstance(data, GenericStatementSink):
        statements = cast(Generator[Quad], data.store)
        graphs = split_to_graphs(statements)
    elif iter(data):
        statements = data
        graphs = split_to_graphs(statements)

    for graph in graphs:
        yield from stream.graph(graph_id=graph.identifier, graph=graph)

    if frame := stream.flow.frame_from_dataset():
        yield frame
    if stream.stream_types.flat and (frame := stream.flow.to_stream_frame()):
        yield frame


def split_to_graphs(data: Generator[Quad]) -> Generator[GenericStatementSink]:
    """
    Split a generator of quads to graphs.

    Notes:
        New graph is generated by
        iterating over statements and yielding one new GenericStatementSink
        per a sequence of quads with the same g term.

    Args:
        data (Generator[Quad]): generator of quads

    Yields:
        Generator[GenericStatementSink]: generator of GenericStatementSinks,
        each having triples in store and identifier set.

    """
    current_g: Node | None = None
    current_sink: GenericStatementSink | None = None
    for statement in data:
        if current_g != statement.g:
            if current_sink is not None:
                yield current_sink

            current_g = statement.g
            current_sink = GenericStatementSink(identifier=current_g)

        assert current_sink is not None
        current_sink.add(Triple(statement.s, statement.p, statement.o))

    if current_sink is not None:
        yield current_sink


def guess_options(sink: GenericStatementSink) -> SerializerOptions:
    """Guess the serializer options based on the store type."""
    logical_type = (
        jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES
        if sink.is_triples_sink
        else jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS
    )
    return SerializerOptions(logical_type=logical_type)


def guess_stream(options: SerializerOptions, sink: GenericStatementSink) -> Stream:
    """
    Return an appropriate stream implementation for the given options.

    Notes: if base(!) logical type is GRAPHS and sink.is_triples_sink is false,
        initializes TripleStream
    """
    stream_cls: type[Stream]
    if (
        options.logical_type % 10
    ) != jelly.LOGICAL_STREAM_TYPE_GRAPHS and not sink.is_triples_sink:
        stream_cls = QuadStream
    else:
        stream_cls = TripleStream
    if options is not None:
        lookup_preset = options.lookup_preset
    return stream_cls(
        encoder=GenericSinkTermEncoder(lookup_preset=lookup_preset),
        options=options,
    )


def grouped_stream_to_frames(
    sink_generator: Generator[GenericStatementSink],
    options: SerializerOptions | None = None,
) -> Generator[jelly.RdfStreamFrame]:
    """
    Transform multiple GenericStatementSinks into Jelly frames.

    Notes:
        One frame per GenericStatementSink.

    Note: options are guessed if not provided.

    Args:
        sink_generator (Generator[GenericStatementSink]): Generator of
            GenericStatementSink to transform.
        options (SerializerOptions | None, optional): stream options to use.
            Options are guessed based on the sink store type. Defaults to None.

    Yields:
        Generator[jelly.RdfStreamFrame]: produced Jelly frames

    """
    stream = None
    for sink in sink_generator:
        if not stream:
            if options is None:
                options = guess_options(sink)
            stream = guess_stream(options, sink)
        yield from stream_frames(stream, sink)


def grouped_stream_to_file(
    stream: Generator[GenericStatementSink],
    output_file: IO[bytes],
    **kwargs: Any,
) -> None:
    """
    Write stream of GenericStatementSink to a binary file.

    Args:
        stream (Generator[GenericStatementSink]): Generator of
            GenericStatementSink to serialize.
        output_file (IO[bytes]): output buffered writer.
        **kwargs (Any): options to pass to stream.

    """
    for frame in grouped_stream_to_frames(stream, **kwargs):
        write_delimited(frame, output_file)


def flat_stream_to_frames(
    statements: Generator[Triple | Quad],
    options: SerializerOptions | None = None,
) -> Generator[jelly.RdfStreamFrame]:
    """
    Serialize a stream of raw GenericStatementSink's triples or quads into Jelly frames.

    Args:
        statements (Generator[Triple | Quad]):
          s/p/o triples or s/p/o/g quads to serialize.
        options (SerializerOptions | None, optional):
            if omitted, guessed based on the first tuple.

    Yields:
        Generator[jelly.RdfStreamFrame]: generated frames.

    """
    first = next(statements, None)
    if first is None:
        return

    sink = GenericStatementSink()
    sink.add(first)
    if options is None:
        options = guess_options(sink)
    stream = guess_stream(options, sink)

    combined: Generator[Triple | Quad] | GenericStatementSink = (
        item for item in chain([first], statements)
    )

    yield from stream_frames(stream, combined)


def flat_stream_to_file(
    statements: Generator[Triple | Quad],
    output_file: IO[bytes],
    options: SerializerOptions | None = None,
) -> None:
    """
    Write Triple or Quad events to a binary file.

    Args:
        statements (Generator[Triple | Quad]): statements to serialize.
        output_file (IO[bytes]): output buffered writer.
        options (SerializerOptions | None, optional): stream options.

    """
    for frame in flat_stream_to_frames(statements, options):
        write_delimited(frame, output_file)



import importlib
import pathlib
import runpy
import urllib.request
from typing import IO

import pytest

BASE_DIR = pathlib.Path(__file__).parent


def resolve_scripts_dir(name: str) -> pathlib.Path:
    path = BASE_DIR / name
    if path.is_file():
        return (BASE_DIR / pathlib.Path(path.read_text().strip())).resolve()
    if path.is_dir():
        return path.resolve()
    raise FileNotFoundError(path)


SCRIPTS_RDFLIB = resolve_scripts_dir("examples/examples_rdflib")
SCRIPTS_GENERIC = resolve_scripts_dir("examples/examples_generic")
SCRIPTS_RDFLIB_CASE = ["rdflib/03_parse_autodetect.py"]

example_scripts = [
    pytest.param("rdflib", p, id=f"rdflib/{p.name}")
    for p in sorted(SCRIPTS_RDFLIB.glob("*.py"))
] + [
    pytest.param("generic", p, id=f"generic/{p.name}")
    for p in sorted(SCRIPTS_GENERIC.glob("*.py"))
]


@pytest.mark.parametrize(
    ("set_name", "script"), example_scripts, ids=lambda p: f"{p[0]}/{p[1].name}"
)
def test_examples(
    set_name: str, script: pathlib.Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    temp_dir = pathlib.Path(__file__, "..", "temp").resolve()

    if f"{set_name}/{script.name}" in SCRIPTS_RDFLIB_CASE:
        # required to mock __init__ for 03_parse_autodetect in order to work
        import pyjelly.options as _opts

        monkeypatch.setattr(_opts, "INTEGRATION_SIDE_EFFECTS", True, raising=False)
        pytest.importorskip("pyjelly.integrations.rdflib")
        import pyjelly.integrations.rdflib as _integration

        importlib.reload(_integration)

    # Run the examples in a temporary directory to avoid polluting the repository
    monkeypatch.chdir(temp_dir)
    # Mock HTTP requests to avoid network calls during tests
    monkeypatch.setattr(urllib.request, "urlopen", urlopen_mock)
    runpy.run_path(str(script))


def urlopen_mock(url: str) -> IO[bytes]:
    response_file = ""
    mode = "rb"
    if url.endswith(".jelly.gz"):
        response_file = "sample.jelly.gz"
    elif url.endswith(".jelly"):
        response_file = "sample.jelly"
    elif url.endswith(".gz"):
        response_file = "sample.nt.gz"
    else:
        response_file = "sample.nt"
        mode = "r"
    # ruff: noqa: PTH123
    return open("../example_data/" + response_file, mode=mode)



from __future__ import annotations

from pathlib import Path
from unittest.mock import patch

import pytest
from rdflib import Dataset, Graph, Literal, Node
from rdflib.graph import DATASET_DEFAULT_GRAPH_ID
from rdflib.plugins.serializers.nt import _quoteLiteral

from pyjelly.integrations.generic.parse import (
    parse_jelly_grouped as generic_parse_jelly_grouped,
)
from pyjelly.integrations.rdflib.parse import parse_jelly_grouped
from tests.meta import (
    RDF_FROM_JELLY_TESTS_DIR,
    TEST_OUTPUTS_DIR,
)
from tests.utils.generic_sink_test_serializer import GenericSinkSerializer
from tests.utils.ordered_memory import OrderedMemory
from tests.utils.rdf_test_cases import (
    GeneralizedTestCasesDir,
    PhysicalTypeTestCasesDir,
    RDFStarGeneralizedTestCasesDir,
    RDFStarTestCasesDir,
    id_from_path,
    jelly_validate,
    needs_jelly_cli,
    walk_directories,
)


def _new_nq_row(triple: tuple[Node, Node, Node], context: Graph) -> str:
    template = "%s " * (3 + (context != DATASET_DEFAULT_GRAPH_ID)) + ".\n"
    args = (
        triple[0].n3(),
        triple[1].n3(),
        _quoteLiteral(triple[2]) if isinstance(triple[2], Literal) else triple[2].n3(),
        *((context.n3(),) if context != DATASET_DEFAULT_GRAPH_ID else ()),
    )
    return template % args


workaround_rdflib_serializes_default_graph_id = patch(
    "rdflib.plugins.serializers.nquads._nq_row",
    new=_new_nq_row,
)


workaround_rdflib_serializes_default_graph_id.start()


@needs_jelly_cli
@walk_directories(
    RDF_FROM_JELLY_TESTS_DIR / PhysicalTypeTestCasesDir.TRIPLES,
    RDF_FROM_JELLY_TESTS_DIR / PhysicalTypeTestCasesDir.QUADS,
    RDF_FROM_JELLY_TESTS_DIR / PhysicalTypeTestCasesDir.GRAPHS,
    glob="pos_*",
)
def test_parses(path: Path) -> None:
    input_filename = path / "in.jelly"
    test_id = id_from_path(path)
    output_dir = TEST_OUTPUTS_DIR / test_id
    output_dir.mkdir(exist_ok=True)
    with input_filename.open("rb") as input_file:
        for frame_no, graph in enumerate(
            parse_jelly_grouped(
                input_file,
                graph_factory=lambda: Graph(store=OrderedMemory()),
                dataset_factory=lambda: Dataset(store=OrderedMemory()),
            )
        ):
            extension = f"n{'quads' if isinstance(graph, Dataset) else 'triples'}"
            output_filename = output_dir / f"out_{frame_no:03}.{extension[:2]}"
            graph.serialize(
                destination=output_filename, encoding="utf-8", format=extension
            )
            jelly_validate(
                input_filename,
                "--compare-ordered",
                "--compare-frame-indices",
                frame_no,
                "--compare-to-rdf-file",
                output_filename,
                hint=f"Test ID: {test_id}, output file: {output_filename}",
            )


@needs_jelly_cli
@walk_directories(
    RDF_FROM_JELLY_TESTS_DIR / PhysicalTypeTestCasesDir.TRIPLES,
    RDF_FROM_JELLY_TESTS_DIR / PhysicalTypeTestCasesDir.QUADS,
    RDF_FROM_JELLY_TESTS_DIR / PhysicalTypeTestCasesDir.GRAPHS,
    glob="pos_*",
)
def test_1_1_parses(path: Path) -> None:
    run_generic_test(path)


@needs_jelly_cli
@walk_directories(
    RDF_FROM_JELLY_TESTS_DIR / GeneralizedTestCasesDir.TRIPLES,
    RDF_FROM_JELLY_TESTS_DIR / GeneralizedTestCasesDir.QUADS,
    RDF_FROM_JELLY_TESTS_DIR / GeneralizedTestCasesDir.GRAPHS,
    glob="pos_*",
)
def test_generalized_parses(path: Path) -> None:
    run_generic_test(path)


@needs_jelly_cli
@walk_directories(
    RDF_FROM_JELLY_TESTS_DIR / RDFStarTestCasesDir.TRIPLES,
    RDF_FROM_JELLY_TESTS_DIR / RDFStarTestCasesDir.QUADS,
    RDF_FROM_JELLY_TESTS_DIR / RDFStarTestCasesDir.GRAPHS,
    glob="pos_*",
)
def test_rdf_star_parses(path: Path) -> None:
    run_generic_test(path)


@needs_jelly_cli
@walk_directories(
    RDF_FROM_JELLY_TESTS_DIR / RDFStarGeneralizedTestCasesDir.TRIPLES,
    RDF_FROM_JELLY_TESTS_DIR / RDFStarGeneralizedTestCasesDir.QUADS,
    RDF_FROM_JELLY_TESTS_DIR / RDFStarGeneralizedTestCasesDir.GRAPHS,
    glob="pos_*",
)
def test_rdf_star_generalized_parses(path: Path) -> None:
    run_generic_test(path)


@needs_jelly_cli
def run_generic_test(path: Path) -> None:
    input_filename = path / "in.jelly"
    test_id = id_from_path(path)
    output_dir = TEST_OUTPUTS_DIR / test_id
    output_dir.mkdir(exist_ok=True)
    with input_filename.open("rb") as input_file:
        for frame_no, graph in enumerate(generic_parse_jelly_grouped(input_file)):
            extension = f"n{'triples' if 'triples' in test_id else 'quads'}"
            output_filename = output_dir / f"out_{frame_no:03}.{extension[:2]}"
            serializer = GenericSinkSerializer(graph)
            serializer.serialize(output_filename=output_filename, encoding="utf-8")
            jelly_validate(
                input_filename,
                "--compare-ordered",
                "--compare-frame-indices",
                frame_no,
                "--compare-to-rdf-file",
                output_filename,
                hint=f"Test ID: {test_id}, output file: {output_filename}",
            )


@needs_jelly_cli
@walk_directories(
    RDF_FROM_JELLY_TESTS_DIR / PhysicalTypeTestCasesDir.TRIPLES,
    RDF_FROM_JELLY_TESTS_DIR / PhysicalTypeTestCasesDir.GRAPHS,
    RDF_FROM_JELLY_TESTS_DIR / PhysicalTypeTestCasesDir.QUADS,
    glob="neg_*",
)
def test_parsing_fails(path: Path) -> None:
    input_filename = str(path / "in.jelly")
    test_id = id_from_path(path)
    output_dir = TEST_OUTPUTS_DIR / test_id
    output_dir.mkdir(exist_ok=True)
    dataset = Dataset(store=OrderedMemory())
    with pytest.raises(Exception):  # TODO: more specific  # noqa: PT011, B017, TD002
        dataset.parse(location=input_filename, format="jelly")


@needs_jelly_cli
def run_generic_fail_test(path: Path) -> None:
    input_filename = path / "in.jelly"
    test_id = id_from_path(path)
    output_dir = TEST_OUTPUTS_DIR / test_id
    output_dir.mkdir(exist_ok=True)

    with (
        pytest.raises(Exception),  # TODO: more specific  # noqa: PT011, B017, TD002
        input_filename.open("rb") as input_file,
    ):
        list(generic_parse_jelly_grouped(input_file))


@needs_jelly_cli
@walk_directories(
    RDF_FROM_JELLY_TESTS_DIR / RDFStarTestCasesDir.TRIPLES,
    RDF_FROM_JELLY_TESTS_DIR / RDFStarTestCasesDir.GRAPHS,
    RDF_FROM_JELLY_TESTS_DIR / RDFStarTestCasesDir.QUADS,
    glob="neg_*",
)
def test_parsing_rdf_star_fails(path: Path) -> None:
    run_generic_fail_test(path)


@needs_jelly_cli
@walk_directories(
    RDF_FROM_JELLY_TESTS_DIR / PhysicalTypeTestCasesDir.TRIPLES,
    RDF_FROM_JELLY_TESTS_DIR / PhysicalTypeTestCasesDir.GRAPHS,
    RDF_FROM_JELLY_TESTS_DIR / PhysicalTypeTestCasesDir.QUADS,
    glob="neg_*",
)
def test_parsing_rdf_1_1_fails(path: Path) -> None:
    run_generic_fail_test(path)


@needs_jelly_cli
@walk_directories(
    RDF_FROM_JELLY_TESTS_DIR / RDFStarGeneralizedTestCasesDir.TRIPLES,
    RDF_FROM_JELLY_TESTS_DIR / RDFStarGeneralizedTestCasesDir.GRAPHS,
    RDF_FROM_JELLY_TESTS_DIR / RDFStarGeneralizedTestCasesDir.QUADS,
    glob="neg_*",
)
def test_parsing_rdf_star_generalized_fails(path: Path) -> None:
    run_generic_fail_test(path)



"""
MkDocs macros plugin configuration for pyjelly documentation.

Defines custom macros used in Markdown docs:
- `python_version()`: current package version (from TAG env or git)
- `proto_version()`: current Jelly proto version (from PROTO_TAG or git)
- `git_link(file)`: deep link to source on GitHub
- `proto_link(page)`: link to W3ID Jelly proto docs
- `python_package_version[_minor]`: current PyPI version(s)

These macros are used in versioned builds and cross-version linking.

See: https://mkdocs-macros-plugin.readthedocs.io/en/latest/
"""

# ruff: noqa: PGH004
# ruff: noqa
import os
import subprocess


def define_env(env):
    # Override for use in local development
    proto_tag_raw = os.environ.get("PROTO_TAG", None)
    if proto_tag_raw is not None:
        print(f"PROTO_TAG env var is set to {proto_tag_raw}")
    else:
        try:
            proto_tag_raw = subprocess.check_output(
                ["git", "describe", "--tags"],
                cwd="submodules/protobuf",
                encoding="utf-8",
            ).strip()
        except subprocess.CalledProcessError as e:
            print("Failed to call git: ", e.returncode, e.stderr)
            raise

    try:
        python_tag_raw = subprocess.check_output(
            ["uv", "version", "--short"], encoding="utf-8"
        ).strip()
    except subprocess.CalledProcessError as e:
        print("Failed to call git: ", e.returncode, e.stderr)
        raise

    tag_env_var = os.environ.get("TAG", "dev")
    if tag_env_var == "dev":
        print("Warning: TAG env var is not set, using dev as default")

    def proto_tag():
        if proto_tag_raw.count("-") > 1:
            if python_version() == "dev":
                print(
                    f"Warning: proto tag ({proto_tag_raw}) contains more than one hyphen, using dev instead"
                )
                return "dev"
            new_tag = "-".join(proto_tag_raw.split("-")[:2])
            print(
                f"Warning: proto tag ({proto_tag_raw}) contains more than one hyphen. Using {new_tag} instead. To fix this, you must update the protobuf_shared submodule to some stable tag."
            )
            return new_tag
        return proto_tag_raw

    @env.macro
    def python_version():
        if tag_env_var == "dev":
            return tag_env_var
        if tag_env_var == "main":
            return "dev"
        return tag_env_var.replace("v", "")

    @env.macro
    def python_package_version():
        """Returns the current python package version, as published to PyPI."""
        v = python_version()
        if v == "dev":
            return python_tag_raw.split("-")[0].replace("v", "")
        return v

    @env.macro
    def python_package_version_minor():
        """Return the current MINOR python package version, as published to PyPI."""
        v = python_package_version()
        return ".".join(v.split(".")[:2]) + ".x"

    @env.macro
    def git_tag():
        return os.environ.get("TAG", "main")

    @env.macro
    def git_link(file: str):
        tag = git_tag()
        return f"https://github.com/Jelly-RDF/pyjelly/blob/{tag}/{file}"

    @env.macro
    def proto_version():
        if python_version() == "dev":
            return "dev"
        tag = proto_tag()
        if "-" in tag:
            tag = tag[: tag.rindex("-")]
        return tag.replace("v", "")

    @env.macro
    def proto_link(page: str = ""):
        version = proto_version()
        return f"https://w3id.org/jelly/{version}/{page}"

    @env.macro
    def code_example(file_name, start=None, end=None):
        try:
            lines = open(f"examples/{file_name}").read().splitlines(keepends=True)
        except:
            return "```python\n\n```"
        start_index = 0 if not start or start < 1 else start - 1
        end_index = len(lines) if not end or end < 1 else end
        if start_index < 0:
            start_index = 0
        if end_index > len(lines):
            end_index = len(lines)
        if start_index >= end_index:
            return "```python\n\n```"
        return "```python\n" + "".join(lines[start_index:end_index]) + "```"

    @env.macro
    def code_example_box(file_name):
        name = f"examples/{file_name}"
        with open(name, "r") as f:
            code = f.read()
        code = code.replace("\n", "\n    ")
        return f"""
??? example "Example: {file_name} (click to expand)"

    **[:octicons-code-24: Source code on GitHub]({git_link(name)})**

    ```python title="{file_name}" linenums="1"
    {code}
    ```
"""

    @env.macro
    def snippet_raw_lines(path, start, end):
        with open(path, "r") as f:
            lines = f.readlines()
        if start < 1:
            start = 1
        if end > len(lines):
            end = len(lines)
        if start > end:
            return (
                f"Error: invalid range {start}-{end} for file with {len(lines)} lines."
            )
        snippet = "".join(lines[start - 1 : end])
        return snippet

    @env.macro
    def snippet_admonition(
        path, start, end, title="Snippet", expanded=True, kind="note"
    ):
        marker = "???+" if expanded else "???"
        snippet = snippet_raw_lines(path, start, end)

        if snippet.startswith("Error:"):
            inner = f"""```text
{snippet}
```"""
        else:
            inner = f"```python\n{snippet}```"
        indented = "\n".join(
            ("    " + line) if line.strip() != "" else "" for line in inner.splitlines()
        )
        return f'{marker} {kind} "{title}"\n{indented}'

    def transform_nav_item(item):
        if list(item.values())[0] == "https://w3id.org/jelly/":
            return {list(item.keys())[0]: proto_link("")}
        return item

    env.conf["nav"] = [transform_nav_item(item) for item in env.conf["nav"]]



from __future__ import annotations

from collections import deque
from dataclasses import dataclass

from pyjelly.errors import JellyAssertionError, JellyConformanceError
from pyjelly.options import MAX_LOOKUP_SIZE


@dataclass
class LookupDecoder:
    """
    Shared base for RDF lookup encoders using Jelly compression.

    Tracks the last assigned and last reused index.

    Parameters
    ----------
    lookup_size
        Maximum lookup size.

    """

    last_assigned_index: int
    last_reused_index: int

    def __init__(self, *, lookup_size: int) -> None:
        if lookup_size > MAX_LOOKUP_SIZE:
            msg = f"lookup size must be less than {MAX_LOOKUP_SIZE}"
            raise JellyAssertionError(msg)
        self.lookup_size = lookup_size
        placeholders = (None,) * lookup_size
        self.data: deque[str | None] = deque(placeholders, maxlen=lookup_size)
        self.last_assigned_index = 0
        self.last_reused_index = 0

    def assign_entry(self, index: int, value: str) -> None:
        previous_index = self.last_assigned_index
        if index == 0:
            index = previous_index + 1
        assert index > 0
        self.data[index - 1] = value
        self.last_assigned_index = index

    def at(self, index: int) -> str:
        self.last_reused_index = index
        value = self.data[index - 1]
        if value is None:
            msg = f"invalid resolved index {index}"
            raise IndexError(msg)
        return value

    def decode_prefix_term_index(self, index: int) -> str:
        actual_index = index or self.last_reused_index
        if actual_index == 0:
            return ""
        return self.at(actual_index)

    def decode_name_term_index(self, index: int) -> str:
        actual_index = index or self.last_reused_index + 1
        if actual_index == 0:
            msg = "0 is not a valid name term index"
            raise JellyConformanceError(msg)
        return self.at(actual_index)

    def decode_datatype_term_index(self, index: int) -> str | None:
        if index == 0:
            msg = "0 is not a valid datatype term index"
            raise JellyConformanceError(msg)
        return self.at(index)



from pyjelly.integrations.rdflib.serialize import grouped_stream_to_file

from rdflib import Graph, Literal, Namespace
import random


def generate_sample_graphs():
    ex = Namespace("http://example.org/")
    for _ in range(10):
        g = Graph()
        g.add((ex.sensor, ex.temperature, Literal(random.random())))
        g.add((ex.sensor, ex.humidity, Literal(random.random())))
        yield g


output_file_name = "output.jelly"

print(f"Streaming graphs into {output_file_name}…")
sample_graphs = generate_sample_graphs()
with open(output_file_name, "wb") as out_file:
    grouped_stream_to_file(sample_graphs, out_file)
print("All done.")



from __future__ import annotations

from collections import OrderedDict
from dataclasses import dataclass
from typing import final


@final
class Lookup:
    """
    Fixed-size 1-based string-to-index mapping with LRU eviction.

    - Assigns incrementing indices starting from 1.
    - After reaching the maximum size, reuses the existing indices from evicting
      the least-recently-used entries.
    - Index 0 is reserved for delta encoding in Jelly streams.

    To check if a key exists, use `.move(key)` and catch `KeyError`.
    If `KeyError` is raised, the key can be inserted with `.insert(key)`.

    Parameters
    ----------
    max_size
        Maximum number of entries. Zero disables lookup.

    """

    def __init__(self, max_size: int) -> None:
        self.data = OrderedDict[str, int]()
        self.max_size = max_size
        self._evicting = False

    def make_last_to_evict(self, key: str) -> None:
        self.data.move_to_end(key)

    def insert(self, key: str) -> int:
        if not self.max_size:
            msg = "lookup is zero, cannot insert"
            raise IndexError(msg)
        assert key not in self.data, f"key {key!r} already present"
        if self._evicting:
            _, index = self.data.popitem(last=False)
            self.data[key] = index
        else:
            index = len(self.data) + 1
            self.data[key] = index
            self._evicting = index == self.max_size
        return index

    def __repr__(self) -> str:
        max_size, data = self.max_size, self.data
        return f"Lookup({max_size=!r}, {data=!r})"


@dataclass
class LookupEncoder:
    """
    Shared base for RDF lookup encoders using Jelly compression.

    Tracks the last assigned and last reused index.

    Parameters
    ----------
    lookup_size
        Maximum lookup size.

    """

    last_assigned_index: int
    last_reused_index: int

    def __init__(self, *, lookup_size: int) -> None:
        self.lookup = Lookup(max_size=lookup_size)
        self.last_assigned_index = 0
        self.last_reused_index = 0

    def encode_entry_index(self, key: str) -> int | None:
        """
        Get or assign the index to use in an entry.

        Returns
        -------
        int or None
            - 0 if the new index is sequential (`last_assigned_index + 1`)
            - actual assigned/reused index otherwise
            - None if the key already exists

        If the return value is None, the entry is already in the lookup and does not
        need to be emitted. Any integer value (including 0) means the entry is new
        and should be emitted.

        """
        try:
            self.lookup.make_last_to_evict(key)
            return None  # noqa: TRY300
        except KeyError:
            previous_index = self.last_assigned_index
            index = self.lookup.insert(key)
            self.last_assigned_index = index
            if index == previous_index + 1:
                return 0
            return index

    def encode_term_index(self, value: str) -> int:
        self.lookup.make_last_to_evict(value)
        current_index = self.lookup.data[value]
        self.last_reused_index = current_index
        return current_index

    def encode_prefix_term_index(self, value: str) -> int:
        if self.lookup.max_size == 0:
            return 0
        previous_index = self.last_reused_index
        if not value and previous_index == 0:
            return 0
        current_index = self.encode_term_index(value)
        if previous_index == 0:
            return current_index
        if current_index == previous_index:
            return 0
        return current_index

    def encode_name_term_index(self, value: str) -> int:
        previous_index = self.last_reused_index
        current_index = self.encode_term_index(value)
        if current_index == previous_index + 1:
            return 0
        return current_index

    def encode_datatype_term_index(self, value: str) -> int:
        if self.lookup.max_size == 0:
            return 0
        return self.encode_term_index(value)



from pyjelly.integrations.rdflib.serialize import flat_stream_to_file
from rdflib import Literal, Namespace
import random


# example generator with triples statements
def generate_sample_triples():
    ex = Namespace("http://example.org/")
    for _ in range(10):
        yield (ex.sensor, ex.temperature, Literal(random.random()))


output_file_name = "flat_output.jelly"

print(f"Streaming triples into {output_file_name}…")
sample_triples = generate_sample_triples()
with open(output_file_name, "wb") as out_file:
    flat_stream_to_file(sample_triples, out_file)
print("All done.")



from unittest.mock import patch

import pytest
from inline_snapshot import snapshot
from pytest_subtests import SubTests

from pyjelly.serialize.lookup import LookupEncoder


def test_encode_entry_index() -> None:
    encoder = LookupEncoder(lookup_size=4)

    entry_index = encoder.encode_entry_index("foo")
    assert encoder.lookup.data["foo"] == 1
    # If 0 appears in the first IRI of the stream it MUST be interpreted as 1
    assert entry_index == 0

    entry_index = encoder.encode_entry_index("")
    assert encoder.lookup.data[""] == 2
    # default value of 0 MUST be interpreted as previous_name_id + 1
    assert entry_index == 0

    entry_index = encoder.encode_entry_index("bar")
    assert encoder.lookup.data["bar"] == 3
    assert entry_index == 0

    encoder.last_assigned_index = 10
    entry_index = encoder.encode_entry_index("baz")
    # default value behavior works only if value equals previous + 1
    assert entry_index == snapshot(4)


def test_last_assigned_index() -> None:
    encoder = LookupEncoder(lookup_size=1)

    assert encoder.last_assigned_index == 0

    passthrough = object()

    with patch.object(encoder.lookup, "insert", return_value=passthrough):
        encoder.encode_entry_index("foo")

        assert encoder.last_assigned_index is passthrough


def test_last_reused_index() -> None:
    encoder = LookupEncoder(lookup_size=3)

    assert encoder.last_assigned_index == 0

    encoder.encode_entry_index("foo")

    passthrough = object()
    encoder.lookup.data["foo"] = passthrough  # type: ignore[assignment]
    encoder.encode_term_index("foo")

    assert encoder.last_reused_index is passthrough


def test_encode_term_index() -> None:
    encoder = LookupEncoder(lookup_size=5)

    encoder.encode_entry_index("foo")
    encoder.encode_entry_index("bar")
    encoder.encode_entry_index("biz")
    encoder.encode_entry_index("baz")
    encoder.encode_entry_index("qux")

    assert encoder.encode_term_index("qux") == 5
    assert encoder.encode_term_index("bar") == 2
    assert encoder.encode_term_index("baz") == 4
    assert encoder.encode_term_index("foo") == 1
    assert encoder.encode_term_index("biz") == 3


def test_encode_name_term_index(subtests: SubTests) -> None:
    with subtests.test("lookup size = 3 encodes correctly"):
        encoder = LookupEncoder(lookup_size=3)

        encoder.encode_entry_index("foo")
        encoder.encode_entry_index("bar")

        # Default value of 0 MUST be interpreted as previous_name_id + 1
        previous_id = encoder.last_reused_index
        assert encoder.lookup.data["foo"] == previous_id + 1
        assert encoder.encode_name_term_index("foo") == 0

        previous_id = encoder.last_reused_index
        assert encoder.lookup.data["bar"] == previous_id + 1
        assert encoder.encode_name_term_index("bar") == 0

        encoder.encode_entry_index("baz")
        previous_id = encoder.last_reused_index
        assert encoder.lookup.data["baz"] == previous_id + 1
        assert encoder.encode_name_term_index("baz") == 0

        # Test all non-specialized cases are delegated to previously tested
        # encode_term_index
        passthrough = object()

        with patch.object(encoder, "encode_term_index", return_value=passthrough):
            assert encoder.encode_name_term_index("baz") is passthrough

    # [max_name_table_size] (...) MUST be set to a value greater than or equal to 8.
    with subtests.test("lookup size = 0 fails"):
        encoder = LookupEncoder(lookup_size=0)

        with pytest.raises(KeyError):
            encoder.encode_name_term_index("foo")

        with pytest.raises(KeyError):
            encoder.encode_name_term_index("bar")


def test_encode_prefix_term_index(subtests: SubTests) -> None:
    with subtests.test("empty prefix encodes 0 at first"):
        encoder = LookupEncoder(lookup_size=3)
        assert encoder.encode_prefix_term_index("") == 0
        assert not encoder.lookup.data

    with subtests.test("empty prefix encoded after non-empty prefix"):
        encoder = LookupEncoder(lookup_size=3)
        encoder.encode_entry_index("foo")
        encoder.encode_entry_index("")

        assert encoder.encode_prefix_term_index("foo") == 1
        assert encoder.encode_prefix_term_index("") == 2

    with subtests.test("lookup size = 3 encodes correctly"):
        encoder = LookupEncoder(lookup_size=3)

        encoder.encode_entry_index("foo")
        encoder.encode_entry_index("bar")

        # The default value of 0 MUST be interpreted as the same value
        # as in the last explicitly specified (non-zero) prefix identifier
        assert encoder.encode_prefix_term_index("foo") == 1
        assert encoder.encode_prefix_term_index("foo") == 0
        assert encoder.encode_prefix_term_index("bar") == 2
        assert encoder.encode_prefix_term_index("bar") == 0

        encoder.encode_entry_index("baz")
        assert encoder.encode_prefix_term_index("baz") == 3
        assert encoder.encode_prefix_term_index("baz") == 0

        encoder.encode_entry_index("qux")

        # Test all non-specialized cases are delegated to previously tested
        # encode_term_index
        passthrough = object()

        with patch.object(encoder, "encode_term_index", return_value=passthrough):
            assert encoder.encode_prefix_term_index("baz") is passthrough

    with subtests.test("lookup size = 0 always encodes 0"):
        encoder = LookupEncoder(lookup_size=0)
        assert encoder.encode_prefix_term_index("foo") == 0
        assert encoder.encode_prefix_term_index("bar") == 0

        # If the [max_prefix_table_size] field is set to 0, the prefix lookup
        # MUST NOT be used in the stream
        with patch.object(encoder, "encode_term_index") as mock:
            mock.assert_not_called()


def test_encode_datatype_term_index(subtests: SubTests) -> None:
    with subtests.test("lookup size = 3 encodes correctly"):
        encoder = LookupEncoder(lookup_size=3)
        # Test all cases are delegated to previously tested encode_term_index
        passthrough = object()

        with patch.object(encoder, "encode_term_index", return_value=passthrough):
            assert encoder.encode_datatype_term_index("foo") is passthrough

    with subtests.test("lookup size = 0 always encodes 0"):
        encoder = LookupEncoder(lookup_size=0)
        assert encoder.encode_datatype_term_index("foo") == 0
        assert encoder.encode_datatype_term_index("bar") == 0

        # If the [max_datatype_table_size] field is set to 0, the datatype lookup
        # MUST NOT be used in the stream
        with patch.object(encoder, "encode_term_index") as mock:
            encoder.encode_datatype_term_index("foo")
            mock.assert_not_called()



"""pyjelly CLI with RDFLib backend for tests."""

from __future__ import annotations

import argparse
from pathlib import Path

import rdflib

from pyjelly.integrations.generic.generic_sink import GenericStatementSink
from pyjelly.integrations.generic.serialize import GenericSinkTermEncoder
from pyjelly.integrations.generic.serialize import (
    stream_frames as generic_stream_frames,
)
from pyjelly.integrations.rdflib.serialize import stream_frames
from pyjelly.parse.decode import ParserOptions
from pyjelly.parse.ioutils import get_options_and_frames
from pyjelly.serialize.ioutils import write_delimited
from pyjelly.serialize.streams import SerializerOptions, stream_for_type
from tests.utils.generic_sink_test_parser import GenericSinkParser
from tests.utils.ordered_memory import OrderedMemory


def write_dataset(
    filenames: list[str | Path],
    out_filename: str | Path,
    options: str | ParserOptions | Path | None = None,
) -> None:
    if not isinstance(options, ParserOptions):
        options = get_options_from(options)
    assert options
    stream = stream_for_type(options.stream_types.physical_type).for_rdflib(
        SerializerOptions(
            logical_type=options.stream_types.logical_type,
            lookup_preset=options.lookup_preset,
            params=options.params,
        )
    )
    with Path(out_filename).open("wb") as file:
        sink: rdflib.Dataset | rdflib.Graph
        for filename in map(str, filenames):
            if filename.endswith(".nq"):
                sink = rdflib.Dataset(store=OrderedMemory())
                sink.parse(location=filename)
            else:
                sink = rdflib.Graph(identifier=filename, store=OrderedMemory())
                sink.parse(location=filename)
            if frames := next(stream_frames(stream, sink)):
                write_delimited(frames, file)


def write_graph(
    filename: str | Path,
    *,
    out_filename: str | Path,
    options: str | ParserOptions | Path | None = None,
) -> None:
    if not isinstance(options, ParserOptions):
        options = get_options_from(options)
    assert options
    graph = rdflib.Graph(store=OrderedMemory())
    graph.parse(location=str(filename))

    stream = stream_for_type(options.stream_types.physical_type).for_rdflib(
        SerializerOptions(
            lookup_preset=options.lookup_preset,
            logical_type=options.stream_types.logical_type,
            params=options.params,
        )
    )
    with Path(out_filename).open("wb") as file:
        graph.serialize(
            destination=file,
            format="jelly",
            stream=stream,
        )


def get_options_from(
    options_filename: str | Path | None = None,
) -> ParserOptions | None:
    if options_filename is not None:
        with Path(options_filename).open("rb") as options_file:
            options, _ = get_options_and_frames(options_file)
    else:
        options = None
    return options


def write_graph_or_dataset(
    first: str | Path,
    *extra: str | Path,
    out_filename: str | Path = "out.jelly",
    options: str | Path | ParserOptions | None = None,
) -> None:
    if str(first).endswith(".nq") or extra:
        write_dataset([first, *extra], out_filename=out_filename, options=options)
    else:
        write_graph(first, out_filename=out_filename, options=options)


def write_generic_sink(
    first: str | Path,
    *extra: str | Path,
    out_filename: str | Path = "out.jelly",
    options: str | Path | ParserOptions | None = None,
) -> None:
    if not isinstance(options, ParserOptions):
        options = get_options_from(options)
    assert options
    if options is not None:
        lookup_preset = options.lookup_preset
    stream_class = stream_for_type(options.stream_types.physical_type)
    stream = stream_class(
        encoder=GenericSinkTermEncoder(lookup_preset=lookup_preset),
        options=SerializerOptions(
            lookup_preset=options.lookup_preset,
            logical_type=options.stream_types.logical_type,
            params=options.params,
        ),
    )
    filenames = [first, *extra]
    with Path(out_filename).open("wb") as file:
        sink: GenericStatementSink
        for filename in map(str, filenames):
            sink = GenericStatementSink()
            sink_parser = GenericSinkParser(sink)
            sink_parser.parse(Path(filename))
            if frames := next(generic_stream_frames(stream, sink)):
                write_delimited(frames, file)


if __name__ == "__main__":
    cli = argparse.ArgumentParser()
    cli.add_argument("first", type=str)
    cli.add_argument("extra", nargs="*", type=str)
    cli.add_argument("out", nargs="?", default="out.jelly", type=str)
    cli.add_argument("--options-from", type=str)
    args = cli.parse_args()
    write_graph_or_dataset(
        args.first,
        *args.extra,
        out_filename=args.out,
        options=args.options_from,
    )



import io

import pytest
from rdflib import Dataset, Graph
from rdflib.graph import DATASET_DEFAULT_GRAPH_ID

from pyjelly import jelly
from pyjelly.parse.ioutils import get_options_and_frames
from pyjelly.serialize.streams import GraphStream, QuadStream, SerializerOptions, Stream


def test_flat_triples() -> None:
    g_out = Graph()
    g_out.parse(source="tests/e2e_test_cases/triples_rdf_1_1/nt-syntax-subm-01.nt")

    out = g_out.serialize(encoding="jelly", format="jelly")
    triples_out = set(g_out)
    assert len(triples_out) > 0

    g_in = Graph()
    g_in.parse(out, format="jelly")

    triples_in = set(g_in)

    assert len(triples_out) == len(triples_in)
    assert triples_in == triples_out

    options, _ = get_options_and_frames(io.BytesIO(out))
    assert options.stream_types.physical_type == jelly.PHYSICAL_STREAM_TYPE_TRIPLES
    assert options.stream_types.logical_type == jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES

    ds_in = Dataset()
    ds_in.parse(out, format="jelly")

    quads_in = set(ds_in)
    assert quads_in == {(*triple, DATASET_DEFAULT_GRAPH_ID) for triple in triples_in}


@pytest.mark.parametrize("stream_class", [QuadStream, GraphStream])
def test_flat_quads(stream_class: type[Stream]) -> None:
    ds_out = Dataset()
    ds_out.parse(source="tests/e2e_test_cases/quads_rdf_1_1/weather-quads.nq")

    stream = stream_class.for_rdflib()

    out = ds_out.serialize(encoding="jelly", format="jelly", stream=stream)
    quads_out = set(ds_out)
    assert len(quads_out) > 0

    options, _ = get_options_and_frames(io.BytesIO(out))
    assert options.stream_types.physical_type == stream_class.physical_type
    assert options.stream_types.logical_type == jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS

    ds_in = Dataset()
    ds_in.parse(out, format="jelly")

    quads_in = set(ds_in)
    assert quads_in == quads_out


@pytest.mark.skip
def test_graphs() -> None:
    # TODO(Nastya): rewrite or remove
    options = SerializerOptions(logical_type=jelly.LOGICAL_STREAM_TYPE_GRAPHS)

    ds_out = Dataset()
    g1_out = Graph(identifier="foaf")
    g1_out.parse(source="tests/e2e_test_cases/triples_rdf_1_1/nt-syntax-subm-01.nt")
    g2_out = Graph(identifier="test")
    g2_out.parse(source="tests/e2e_test_cases/triples_rdf_1_1/p2_ontology.nt")
    ds_out.add_graph(g1_out)
    ds_out.add_graph(g2_out)

    out = ds_out.serialize(options=options, encoding="jelly", format="jelly")

    options_out, _ = get_options_and_frames(io.BytesIO(out))
    assert options_out.stream_types.physical_type == jelly.PHYSICAL_STREAM_TYPE_TRIPLES
    assert options_out.stream_types.logical_type == jelly.LOGICAL_STREAM_TYPE_GRAPHS

    graphs_out = sorted(ds_out.graphs(), key=len)

    ds_in = Dataset()
    ds_in.parse(out, format="jelly")

    graphs_in = sorted(ds_in.graphs(), key=len)

    for g_out, g_in in zip(graphs_out, graphs_in):
        assert len(g_out) == len(g_in)
        assert set(g_out) == set(g_in)



import networkx as nx
from rdflib import Graph, URIRef, Namespace
from rdflib.namespace import RDF

# An example NetworkX graph
nx_g = nx.Graph()
nx_g.add_node("http://example.org/A")
nx_g.add_node("http://example.org/B")
nx_g.add_edge("http://example.org/A", "http://example.org/B")

# We define RDFLib graph for further conversion
rdf_g = Graph()

# Example namespace
ex = Namespace("http://example.org/ns#")

# Add triples through node information in NetworkX graph
for node_uri, data in nx_g.nodes(data=True):
    subj = URIRef(node_uri)
    rdf_g.add((subj, RDF.type, ex.Node))

# Add triples through edge information in NetworkX graph
for u, v, attr in nx_g.edges(data=True):
    rdf_g.add((URIRef(u), ex.connectedTo, URIRef(v)))

# Serialize graph into a .jelly file
rdf_g.serialize(destination="networkx_graph.jelly", format="jelly")
print("All done.")



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from collections.abc import Iterable, Iterator, Sequence
from enum import Enum, auto
from typing import Any, ClassVar, NamedTuple
from typing_extensions import Never

from pyjelly import jelly
from pyjelly.options import LookupPreset, StreamParameters, StreamTypes
from pyjelly.parse.lookup import LookupDecoder


class ParsingMode(Enum):
    """
    Specifies how jelly frames should be treated.

    Modes:
    FLAT
        Yield all frames as one Graph or Dataset.
    GROUPED
        Yield one Graph/Dataset per frame (grouped parsing).
    """

    FLAT = auto()
    GROUPED = auto()


class ParserOptions(NamedTuple):
    stream_types: StreamTypes
    lookup_preset: LookupPreset
    params: StreamParameters


def options_from_frame(
    frame: jelly.RdfStreamFrame,
    *,
    delimited: bool,
) -> ParserOptions:
    """
    Fill stream options based on the options row.

    Notes:
        generalized_statements, rdf_star, and namespace declarations
        are set to false by default

    Args:
        frame (jelly.RdfStreamFrame): first non-empty frame from the stream
        delimited (bool): derived delimited flag

    Returns:
        ParserOptions: filled options with types/lookups/stream parameters information

    """
    row = frame.rows[0]
    options = row.options
    return ParserOptions(
        stream_types=StreamTypes(
            physical_type=options.physical_type,
            logical_type=options.logical_type,
        ),
        lookup_preset=LookupPreset(
            max_names=options.max_name_table_size,
            max_prefixes=options.max_prefix_table_size,
            max_datatypes=options.max_datatype_table_size,
        ),
        params=StreamParameters(
            stream_name=options.stream_name,
            generalized_statements=options.generalized_statements,
            rdf_star=options.rdf_star,
            version=options.version,
            delimited=delimited,
        ),
    )


def _adapter_missing(feature: str, *, stream_types: StreamTypes) -> Never:
    physical_type_name = jelly.PhysicalStreamType.Name(stream_types.physical_type)
    logical_type_name = jelly.LogicalStreamType.Name(stream_types.logical_type)
    msg = (
        f"adapter with {physical_type_name} and {logical_type_name} "
        f"does not implement {feature}"
    )
    raise NotImplementedError(msg)


class Adapter(metaclass=ABCMeta):
    def __init__(
        self, options: ParserOptions, parsing_mode: ParsingMode = ParsingMode.FLAT
    ) -> None:
        self.options = options
        self.parsing_mode = parsing_mode

    # Obligatory abstract methods--all adapters must implement these
    @abstractmethod
    def iri(self, iri: str) -> Any:
        raise NotImplementedError

    @abstractmethod
    def default_graph(self) -> Any:
        raise NotImplementedError

    @abstractmethod
    def bnode(self, bnode: str) -> Any:
        raise NotImplementedError

    @abstractmethod
    def literal(
        self,
        lex: str,
        language: str | None = None,
        datatype: str | None = None,
    ) -> Any:
        raise NotImplementedError

    # Optional abstract methods--not required to be implemented by all adapters
    def triple(self, terms: Iterable[Any]) -> Any:  # noqa: ARG002
        _adapter_missing("decoding triples", stream_types=self.options.stream_types)

    def quad(self, terms: Iterable[Any]) -> Any:  # noqa: ARG002
        _adapter_missing("decoding quads", stream_types=self.options.stream_types)

    def graph_start(self, graph_id: Any) -> Any:  # noqa: ARG002
        _adapter_missing(
            "decoding graph start markers", stream_types=self.options.stream_types
        )

    def graph_end(self) -> Any:
        _adapter_missing(
            "decoding graph end markers", stream_types=self.options.stream_types
        )

    def namespace_declaration(self, name: str, iri: str) -> Any:  # noqa: ARG002
        _adapter_missing(
            "decoding namespace declarations",
            stream_types=self.options.stream_types,
        )

    def quoted_triple(self, terms: Iterable[Any]) -> Any:  # noqa: ARG002
        _adapter_missing(
            "decoding quoted triple", stream_types=self.options.stream_types
        )

    def frame(self) -> Any:
        return None


class Decoder:
    def __init__(self, adapter: Adapter) -> None:
        """
        Initialize decoder.

        Initializes decoder with a lookup tables with preset sizes,
        integration-dependent adapter and empty repeated terms dictionary.

        Args:
            adapter (Adapter): integration-dependent adapter that specifies terms
            conversion to specific objects, framing,
            namespace declarations, and graphs/datasets forming.

        """
        self.adapter = adapter
        self.names = LookupDecoder(lookup_size=self.options.lookup_preset.max_names)
        self.prefixes = LookupDecoder(
            lookup_size=self.options.lookup_preset.max_prefixes
        )
        self.datatypes = LookupDecoder(
            lookup_size=self.options.lookup_preset.max_datatypes
        )
        self.repeated_terms: dict[str, jelly.RdfIri | str | jelly.RdfLiteral] = {}

    @property
    def options(self) -> ParserOptions:
        return self.adapter.options

    def iter_rows(self, frame: jelly.RdfStreamFrame) -> Iterator[Any]:
        """
        Iterate through rows in the frame.

        Args:
            frame (jelly.RdfStreamFrame): jelly frame
        Yields:
            Iterator[Any]: decoded rows

        """
        for row_owner in frame.rows:
            row = getattr(row_owner, row_owner.WhichOneof("row"))
            decoded_row = self.decode_row(row)
            if isinstance(
                row, (jelly.RdfTriple, jelly.RdfQuad, jelly.RdfNamespaceDeclaration)
            ):
                yield decoded_row

    def decode_row(self, row: Any) -> Any | None:
        """
        Decode a row based on its type.

        Notes: uses custom adapters to decode triples/quads, namespace declarations,
               graph start/end.

        Args:
            row (Any): protobuf row message

        Raises:
            TypeError: raises error if this type of protobuf message does not have
                       a respective handler

        Returns:
            Any | None: decoded row -
                        result from calling decode_row (row type appropriate handler)

        """
        try:
            decode_row = self.row_handlers[type(row)]
        except KeyError:
            msg = f"decoder not implemented for {type(row)}"
            raise TypeError(msg) from None
        return decode_row(self, row)

    def validate_stream_options(self, options: jelly.RdfStreamOptions) -> None:
        stream_types, lookup_preset, params = self.options
        assert stream_types.physical_type == options.physical_type
        assert stream_types.logical_type == options.logical_type
        assert params.stream_name == options.stream_name
        assert params.version >= options.version
        assert lookup_preset.max_prefixes == options.max_prefix_table_size
        assert lookup_preset.max_datatypes == options.max_datatype_table_size
        assert lookup_preset.max_names == options.max_name_table_size

    def ingest_prefix_entry(self, entry: jelly.RdfPrefixEntry) -> None:
        """
        Update prefix lookup table based on the table entry.

        Args:
            entry (jelly.RdfPrefixEntry): prefix message, containing id and value

        """
        self.prefixes.assign_entry(index=entry.id, value=entry.value)

    def ingest_name_entry(self, entry: jelly.RdfNameEntry) -> None:
        """
        Update name lookup table based on the table entry.

        Args:
            entry (jelly.RdfNameEntry): name message, containing id and value

        """
        self.names.assign_entry(index=entry.id, value=entry.value)

    def ingest_datatype_entry(self, entry: jelly.RdfDatatypeEntry) -> None:
        """
        Update datatype lookup table based on the table entry.

        Args:
            entry (jelly.RdfDatatypeEntry): name message, containing id and value

        """
        self.datatypes.assign_entry(index=entry.id, value=entry.value)

    def decode_term(self, term: Any) -> Any:
        """
        Decode a term based on its type: IRI/literal/BN/default graph.

        Notes: requires a custom adapter with implemented methods for terms decoding.

        Args:
            term (Any): IRI/literal/BN(string)/Default graph message

        Raises:
            TypeError: raises error if no handler for the term is found

        Returns:
            Any: decoded term (currently, rdflib objects, e.g., rdflib.term.URIRef)

        """
        try:
            decode_term = self.term_handlers[type(term)]
        except KeyError:
            msg = f"decoder not implemented for {type(term)}"
            raise TypeError(msg) from None
        return decode_term(self, term)

    def decode_iri(self, iri: jelly.RdfIri) -> Any:
        """
        Decode RdfIri message to IRI using a custom adapter.

        Args:
            iri (jelly.RdfIri): RdfIri message

        Returns:
            Any: IRI, based on adapter implementation, e.g., rdflib.term.URIRef

        """
        name = self.names.decode_name_term_index(iri.name_id)
        prefix = self.prefixes.decode_prefix_term_index(iri.prefix_id)
        return self.adapter.iri(iri=prefix + name)

    def decode_default_graph(self, _: jelly.RdfDefaultGraph) -> Any:
        return self.adapter.default_graph()

    def decode_bnode(self, bnode: str) -> Any:
        """
        Decode string message to blank node (BN) using a custom adapter.

        Args:
            bnode (str): blank node id

        Returns:
            Any: blank node object from the custom adapter

        """
        return self.adapter.bnode(bnode)

    def decode_literal(self, literal: jelly.RdfLiteral) -> Any:
        """
        Decode RdfLiteral to literal based on custom adapter implementation.

        Notes: checks for langtag existence;
               for datatype checks for non-zero table size and datatype field presence

        Args:
            literal (jelly.RdfLiteral): RdfLiteral message

        Returns:
            Any: literal returned by the custom adapter

        """
        language = datatype = None
        if literal.langtag:
            language = literal.langtag
        elif self.datatypes.lookup_size and literal.HasField("datatype"):
            datatype = self.datatypes.decode_datatype_term_index(literal.datatype)
        return self.adapter.literal(
            lex=literal.lex,
            language=language,
            datatype=datatype,
        )

    def decode_namespace_declaration(
        self,
        declaration: jelly.RdfNamespaceDeclaration,
    ) -> Any:
        iri = self.decode_iri(declaration.value)
        return self.adapter.namespace_declaration(declaration.name, iri)

    def decode_graph_start(self, graph_start: jelly.RdfGraphStart) -> Any:
        term = getattr(graph_start, graph_start.WhichOneof("graph"))
        return self.adapter.graph_start(self.decode_term(term))

    def decode_graph_end(self, _: jelly.RdfGraphEnd) -> Any:
        return self.adapter.graph_end()

    def decode_statement(
        self,
        statement: jelly.RdfTriple | jelly.RdfQuad,
        oneofs: Sequence[str],
    ) -> Any:
        """
        Decode a triple/quad message.

        Notes: also updates repeated terms dictionary

        Args:
            statement (jelly.RdfTriple | jelly.RdfQuad): triple/quad message
            oneofs (Sequence[str]): terms s/p/o/g(if quads)

        Raises:
            ValueError: if a missing repeated term is encountered

        Returns:
            Any: a list of decoded terms

        """
        terms = []
        for oneof in oneofs:
            field = statement.WhichOneof(oneof)
            if field:
                jelly_term = getattr(statement, field)
                decoded_term = self.decode_term(jelly_term)
                self.repeated_terms[oneof] = decoded_term
            else:
                decoded_term = self.repeated_terms[oneof]
                if decoded_term is None:
                    msg = f"missing repeated term {oneof}"
                    raise ValueError(msg)
            terms.append(decoded_term)
        return terms

    def decode_triple(self, triple: jelly.RdfTriple) -> Any:
        terms = self.decode_statement(triple, ("subject", "predicate", "object"))
        return self.adapter.triple(terms)

    def decode_quoted_triple(self, triple: jelly.RdfTriple) -> Any:
        oneofs: Sequence[str] = ("subject", "predicate", "object")
        terms = []
        for oneof in oneofs:
            field = triple.WhichOneof(oneof)
            if field:
                jelly_term = getattr(triple, field)
                decoded_term = self.decode_term(jelly_term)
            else:
                msg = "repeated terms are not allowed in quoted triples"
                raise ValueError(msg)
            terms.append(decoded_term)
        return self.adapter.quoted_triple(terms)

    def decode_quad(self, quad: jelly.RdfQuad) -> Any:
        terms = self.decode_statement(quad, ("subject", "predicate", "object", "graph"))
        return self.adapter.quad(terms)

    # dispatch by invariant type (no C3 resolution)
    row_handlers: ClassVar = {
        jelly.RdfStreamOptions: validate_stream_options,
        jelly.RdfPrefixEntry: ingest_prefix_entry,
        jelly.RdfNameEntry: ingest_name_entry,
        jelly.RdfDatatypeEntry: ingest_datatype_entry,
        jelly.RdfTriple: decode_triple,
        jelly.RdfQuad: decode_quad,
        jelly.RdfGraphStart: decode_graph_start,
        jelly.RdfGraphEnd: decode_graph_end,
        jelly.RdfNamespaceDeclaration: decode_namespace_declaration,
    }

    term_handlers: ClassVar = {
        jelly.RdfIri: decode_iri,
        str: decode_bnode,
        jelly.RdfLiteral: decode_literal,
        jelly.RdfDefaultGraph: decode_default_graph,
        jelly.RdfTriple: decode_quoted_triple,
    }



"""pyjelly CLI with RDFLib backend for tests."""

from __future__ import annotations

import argparse
from pathlib import Path

import rdflib


def parse_and_serialize_graph_or_dataset(
    location: str | Path,
    output: str | Path,
) -> None:
    ds = rdflib.Dataset()
    ds.parse(location=str(location), format="jelly")
    ds.serialize(output, format="jelly")


if __name__ == "__main__":
    cli = argparse.ArgumentParser()
    cli.add_argument("location", nargs="?", default="out.jelly", type=str)
    cli.add_argument("output", nargs="?", default="out-parsed.jelly", type=str)
    args = cli.parse_args()
    parse_and_serialize_graph_or_dataset(location=args.location, output=args.output)



from pyjelly.integrations.generic.serialize import flat_stream_to_file
from pyjelly.integrations.generic.generic_sink import Triple, IRI
import random


# Example generator that yields raw triples
def generate_sample_triples():
    content = (
        IRI("http://example.com/sensor"),
        IRI("http://example.com/humidity"),
        IRI(f"http://example.com/{random.random()}"),
    )
    for _ in range(10):
        yield Triple(*content)


output_file = "flat_output.jelly"
print(f"Streaming triples into {output_file}…")
sample_triples = generate_sample_triples()
with open(output_file, "wb") as out_f:
    flat_stream_to_file(sample_triples, out_f)
print("All done.")



import gzip
import urllib.request

from pyjelly.integrations.rdflib.parse import parse_jelly_grouped

# Dataset: Katrina weather measurements (10k graphs)
# Documentation: https://w3id.org/riverbench/datasets/lod-katrina/dev
url = "https://w3id.org/riverbench/datasets/lod-katrina/dev/files/jelly_10K.jelly.gz"

# Load, uncompress .gz file, and pass to Jelly parser, all in a streaming manner
with (
    urllib.request.urlopen(url) as response,
    gzip.open(response) as jelly_stream,
):
    graphs = parse_jelly_grouped(jelly_stream)
    for i, graph in enumerate(graphs):
        print(f"Graph {i} in the stream has {len(graph)} triples")
        # Limit to 50 graphs for demonstration -- the rest will not be parsed
        if i >= 50:
            break



from pyjelly.integrations.generic.generic_sink import (
    GenericStatementSink,
    IRI,
    Triple,
)
from pyjelly.integrations.generic.serialize import grouped_stream_to_file
import random


# Helper function to generate a generator of graphs
def generate_sample_sinks():
    content = (
        IRI("http://example.com/sensor"),
        IRI("http://example.com/humidity"),
        IRI(f"http://example.com/{random.random()}"),
    )
    for i in range(10):
        sink = GenericStatementSink(identifier=f"graph_{i}")
        sink.add(Triple(*content))
        yield sink


output_file = "output.jelly"
print(f"Streaming graphs into {output_file}…")
with open(output_file, "wb") as out_f:
    grouped_stream_to_file(generate_sample_sinks(), out_f)
print("All done.")



from pyjelly.integrations.generic.generic_sink import GenericStatementSink


# Helper generator that streams statements from a Jelly file path
def stream_triples(jelly_path):
    generic_sink = GenericStatementSink()
    with open(jelly_path, "rb") as f:
        generic_sink.parse(f)
    yield from (stmt for stmt in generic_sink)


# Example usage, just printing:
for triple in stream_triples("output.jelly"):
    print(triple)

print("All done.")



from __future__ import annotations

from collections.abc import Generator, Iterable
from dataclasses import dataclass, field
from typing import Any, ClassVar

from pyjelly import jelly
from pyjelly.options import LookupPreset, StreamParameters, StreamTypes
from pyjelly.serialize.encode import (
    Slot,
    TermEncoder,
    encode_namespace_declaration,
    encode_options,
    encode_quad,
    encode_triple,
)
from pyjelly.serialize.flows import (
    DEFAULT_FRAME_SIZE,
    BoundedFrameFlow,
    FlatQuadsFrameFlow,
    FlatTriplesFrameFlow,
    FrameFlow,
    ManualFrameFlow,
    flow_for_type,
)


@dataclass
class SerializerOptions:
    flow: FrameFlow | None = None
    frame_size: int = DEFAULT_FRAME_SIZE
    logical_type: jelly.LogicalStreamType = jelly.LOGICAL_STREAM_TYPE_UNSPECIFIED
    params: StreamParameters = field(default_factory=StreamParameters)
    lookup_preset: LookupPreset = field(default_factory=LookupPreset)


class Stream:
    physical_type: ClassVar[jelly.PhysicalStreamType]
    default_delimited_flow_class: ClassVar[type[BoundedFrameFlow]]

    def __init__(
        self,
        *,
        encoder: TermEncoder,
        options: SerializerOptions | None = None,
    ) -> None:
        self.encoder = encoder
        if options is None:
            options = SerializerOptions()
        self.options = options
        flow = options.flow
        if flow is None:
            flow = self.infer_flow()
        self.flow = flow
        self.repeated_terms = dict.fromkeys(Slot)
        self.enrolled = False
        self.stream_types = StreamTypes(
            physical_type=self.physical_type,
            logical_type=self.flow.logical_type,
        )

    def infer_flow(self) -> FrameFlow:
        """
        Return flow based on the stream options provided.

        Returns:
            FrameFlow: initialised FrameFlow object.

        """
        flow: FrameFlow
        if self.options.params.delimited:
            if self.options.logical_type != jelly.LOGICAL_STREAM_TYPE_UNSPECIFIED:
                flow_class = flow_for_type(self.options.logical_type)
            else:
                flow_class = self.default_delimited_flow_class

            if self.options.logical_type in (
                jelly.LOGICAL_STREAM_TYPE_FLAT_TRIPLES,
                jelly.LOGICAL_STREAM_TYPE_FLAT_QUADS,
            ):
                flow = flow_class(
                    logical_type=self.options.logical_type,
                    frame_size=self.options.frame_size,
                )
            else:
                flow = flow_class(logical_type=self.options.logical_type)
        else:
            flow = ManualFrameFlow(logical_type=self.options.logical_type)
        return flow

    def enroll(self) -> None:
        """Initialize start of the stream."""
        if not self.enrolled:
            self.stream_options()
            self.enrolled = True

    def stream_options(self) -> None:
        """Encode and append stream options row to the current flow."""
        self.flow.append(
            encode_options(
                stream_types=self.stream_types,
                params=self.options.params,
                lookup_preset=self.options.lookup_preset,
            )
        )

    def namespace_declaration(self, name: str, iri: str) -> None:
        """
        Add namespace declaration to jelly stream.

        Args:
            name (str): namespace prefix label
            iri (str): namespace iri

        """
        rows = encode_namespace_declaration(
            name=name,
            value=iri,
            term_encoder=self.encoder,
        )
        self.flow.extend(rows)

    @classmethod
    def for_rdflib(cls, options: SerializerOptions | None = None) -> Stream:
        """
        Initialize stream with RDFLib encoder.

        Args:
            options (SerializerOptions | None, optional): Stream options.
                Defaults to None.

        Raises:
            TypeError: if Stream is passed, and not a Stream for specific physical type.

        Returns:
            Stream: initialized stream with RDFLib encoder.

        """
        if cls is Stream:
            msg = "Stream is an abstract base class, use a subclass instead"
            raise TypeError(msg)
        from pyjelly.integrations.rdflib.serialize import RDFLibTermEncoder

        lookup_preset: LookupPreset | None = None
        if options is not None:
            lookup_preset = options.lookup_preset
        return cls(
            encoder=RDFLibTermEncoder(lookup_preset=lookup_preset),
            options=options,
        )


def stream_for_type(physical_type: jelly.PhysicalStreamType) -> type[Stream]:
    """
    Give a Stream based on physical type specified.

    Args:
        physical_type (jelly.PhysicalStreamType): jelly stream physical type.

    Raises:
        NotImplementedError: if no stream for requested physical type is available.

    Returns:
        type[Stream]: jelly stream

    """
    try:
        stream_cls = STREAM_DISPATCH[physical_type]
    except KeyError:
        msg = (
            "no stream class for physical type "
            f"{jelly.PhysicalStreamType.Name(physical_type)}"
        )
        raise NotImplementedError(msg) from None
    return stream_cls


class TripleStream(Stream):
    physical_type = jelly.PHYSICAL_STREAM_TYPE_TRIPLES
    default_delimited_flow_class: ClassVar = FlatTriplesFrameFlow

    def triple(self, terms: Iterable[object]) -> jelly.RdfStreamFrame | None:
        """
        Process one triple to Protobuf messages.

        Note:
            Adds new rows to the current flow and returns StreamFrame if
            frame size conditions are met.

        Args:
            terms (Iterable[object]): RDF terms to encode.

        Returns:
            jelly.RdfStreamFrame | None: stream frame if
                flow supports frames slicing and current flow is full

        """
        new_rows = encode_triple(
            terms,
            term_encoder=self.encoder,
            repeated_terms=self.repeated_terms,
        )
        self.flow.extend(new_rows)
        return self.flow.frame_from_bounds()


class QuadStream(Stream):
    physical_type = jelly.PHYSICAL_STREAM_TYPE_QUADS
    default_delimited_flow_class: ClassVar = FlatQuadsFrameFlow

    def quad(self, terms: Iterable[object]) -> jelly.RdfStreamFrame | None:
        """
        Process one quad to Protobuf messages.

        Args:
            terms (Iterable[object]): terms to encode.

        Returns:
            jelly.RdfStreamFrame | None: stream frame if
                flow supports frames slicing and current flow is full

        """
        new_rows = encode_quad(
            terms,
            term_encoder=self.encoder,
            repeated_terms=self.repeated_terms,
        )
        self.flow.extend(new_rows)
        return self.flow.frame_from_bounds()


class GraphStream(TripleStream):
    physical_type = jelly.PHYSICAL_STREAM_TYPE_GRAPHS
    default_delimited_flow_class: ClassVar = FlatQuadsFrameFlow

    def graph(
        self,
        graph_id: object,
        graph: Iterable[Iterable[object]],
    ) -> Generator[jelly.RdfStreamFrame]:
        """
        Process one graph into a sequence of jelly frames.

        Args:
            graph_id (object): graph id (BN, Literal, iri, default)
            graph (Iterable[Iterable[object]]): iterable of triples (graph's content)

        Yields:
            Generator[jelly.RdfStreamFrame]: jelly frames.

        """
        [*graph_rows], graph_node = self.encoder.encode_any(graph_id, Slot.graph)
        kw_name = f"{Slot.graph}_{self.encoder.TERM_ONEOF_NAMES[type(graph_node)]}"
        kws: dict[Any, Any] = {kw_name: graph_node}
        start_row = jelly.RdfStreamRow(graph_start=jelly.RdfGraphStart(**kws))
        graph_rows.append(start_row)
        self.flow.extend(graph_rows)
        for triple in graph:
            if frame := self.triple(triple):  # has frame slicing inside
                yield frame
        end_row = jelly.RdfStreamRow(graph_end=jelly.RdfGraphEnd())
        self.flow.append(end_row)
        if frame := self.flow.frame_from_bounds():
            yield frame


STREAM_DISPATCH: dict[jelly.PhysicalStreamType, type[Stream]] = {
    jelly.PHYSICAL_STREAM_TYPE_TRIPLES: TripleStream,
    jelly.PHYSICAL_STREAM_TYPE_QUADS: QuadStream,
    jelly.PHYSICAL_STREAM_TYPE_GRAPHS: GraphStream,
}



import io

from rdflib import Dataset, Graph

from pyjelly.options import LookupPreset
from pyjelly.serialize.flows import FlatQuadsFrameFlow, FlatTriplesFrameFlow
from pyjelly.serialize.streams import (
    QuadStream,
    SerializerOptions,
    TripleStream,
)
from tests.e2e_tests.ser_des.base_ser_des import (
    BaseSerDes,
    QuadGraphType,
    TripleGraphType,
)
from tests.utils.ordered_memory import OrderedMemory
from tests.utils.rdflib_workarounds import fixup_graph


class RdflibSerDes(BaseSerDes):
    """
    Serialization and deserialization using rdflib.

    Args:
        BaseSerDes (_type_): _description_

    Returns:
        _type_: _description_

    """

    name = "rdflib"

    def __init__(self) -> None:
        super().__init__(name=self.name)

    def read_quads(self, in_bytes: bytes) -> QuadGraphType:
        g = Dataset()
        g.parse(data=in_bytes, format="nquads")
        fixup_graph(g)
        return g

    def write_quads(self, in_graph: QuadGraphType) -> bytes:
        destination = io.BytesIO()
        in_graph.serialize(destination=destination, format="nquads")
        return destination.getvalue()

    def read_quads_jelly(self, in_bytes: bytes) -> QuadGraphType:
        g = Dataset()
        g.parse(data=in_bytes, format="jelly")
        fixup_graph(g)
        return g

    def write_quads_jelly(
        self, in_graph: QuadGraphType, preset: LookupPreset, frame_size: int
    ) -> bytes:
        destination = io.BytesIO()
        options = SerializerOptions(
            flow=FlatQuadsFrameFlow(frame_size=frame_size),
            lookup_preset=preset,
        )
        stream = QuadStream.for_rdflib(options)
        in_graph.serialize(destination=destination, format="jelly", stream=stream)
        return destination.getvalue()

    def read_triples(self, in_bytes: bytes) -> TripleGraphType:
        g = Graph(store=OrderedMemory())
        g.parse(data=in_bytes, format="nt")
        fixup_graph(g)
        return g

    def write_triples(self, in_graph: TripleGraphType) -> bytes:
        destination = io.BytesIO()
        in_graph.serialize(destination=destination, format="nt")
        return destination.getvalue()

    def read_triples_jelly(self, in_bytes: bytes) -> TripleGraphType:
        g = Graph()
        g.parse(data=in_bytes, format="jelly")
        fixup_graph(g)
        return g

    def write_triples_jelly(
        self, in_graph: TripleGraphType, preset: LookupPreset, frame_size: int
    ) -> bytes:
        destination = io.BytesIO()
        options = SerializerOptions(
            flow=FlatTriplesFrameFlow(frame_size=frame_size),
            lookup_preset=preset,
        )
        stream = TripleStream.for_rdflib(options)
        in_graph.serialize(destination=destination, format="jelly", stream=stream)
        return destination.getvalue()



from __future__ import annotations

from pathlib import Path

import pytest

from tests.meta import (
    RDF_TO_JELLY_TESTS_DIR,
    TEST_OUTPUTS_DIR,
)
from tests.serialize import write_generic_sink, write_graph_or_dataset
from tests.utils.rdf_test_cases import (
    GeneralizedTestCasesDir,
    PhysicalTypeTestCasesDir,
    RDFStarGeneralizedTestCasesDir,
    RDFStarTestCasesDir,
    id_from_path,
    jelly_validate,
    needs_jelly_cli,
    walk_directories,
)


@needs_jelly_cli
@walk_directories(
    RDF_TO_JELLY_TESTS_DIR / PhysicalTypeTestCasesDir.TRIPLES,
    RDF_TO_JELLY_TESTS_DIR / PhysicalTypeTestCasesDir.QUADS,
    RDF_TO_JELLY_TESTS_DIR / PhysicalTypeTestCasesDir.GRAPHS,
    glob="pos_*",
)
def test_serializes(path: Path) -> None:
    options_filename = path / "stream_options.jelly"
    input_filenames = tuple(path.glob("in_*"))
    test_id = id_from_path(path)
    actual_out = TEST_OUTPUTS_DIR / f"{test_id}.jelly"

    write_graph_or_dataset(
        *input_filenames,
        options=options_filename,
        out_filename=actual_out,
    )
    for frame_no, input_filename in enumerate(input_filenames):
        jelly_validate(
            actual_out,
            "--compare-ordered",
            "--compare-frame-indices",
            frame_no,
            "--compare-to-rdf-file",
            input_filename,
            "--options-file",
            options_filename,
            hint=f"Test ID: {test_id}, tested file: {input_filename}",
        )


@walk_directories(
    RDF_TO_JELLY_TESTS_DIR / PhysicalTypeTestCasesDir.TRIPLES,
    RDF_TO_JELLY_TESTS_DIR / PhysicalTypeTestCasesDir.QUADS,
    RDF_TO_JELLY_TESTS_DIR / PhysicalTypeTestCasesDir.GRAPHS,
    RDF_TO_JELLY_TESTS_DIR / RDFStarGeneralizedTestCasesDir.TRIPLES,
    RDF_TO_JELLY_TESTS_DIR / RDFStarGeneralizedTestCasesDir.QUADS,
    RDF_TO_JELLY_TESTS_DIR / RDFStarTestCasesDir.TRIPLES,
    RDF_TO_JELLY_TESTS_DIR / RDFStarTestCasesDir.QUADS,
    RDF_TO_JELLY_TESTS_DIR / RDFStarTestCasesDir.GRAPHS,
    RDF_TO_JELLY_TESTS_DIR / GeneralizedTestCasesDir.TRIPLES,
    RDF_TO_JELLY_TESTS_DIR / GeneralizedTestCasesDir.QUADS,
    RDF_TO_JELLY_TESTS_DIR / GeneralizedTestCasesDir.GRAPHS,
    glob="pos_*",
)
@needs_jelly_cli
def test_serializes_generic(path: Path) -> None:
    if path:
        options_filename = path / "stream_options.jelly"
        input_filenames = tuple(path.glob("in_*"))
        test_id = id_from_path(path)
        actual_out = TEST_OUTPUTS_DIR / f"{test_id}.jelly"

        write_generic_sink(
            *input_filenames,
            options=options_filename,
            out_filename=actual_out,
        )
        for frame_no, input_filename in enumerate(input_filenames):
            jelly_validate(
                actual_out,
                "--compare-ordered",
                "--compare-frame-indices",
                frame_no,
                "--compare-to-rdf-file",
                input_filename,
                "--options-file",
                options_filename,
                hint=f"Test ID: {test_id}, tested file: {input_filename}",
            )


@needs_jelly_cli
@walk_directories(
    RDF_TO_JELLY_TESTS_DIR / PhysicalTypeTestCasesDir.TRIPLES,
    RDF_TO_JELLY_TESTS_DIR / PhysicalTypeTestCasesDir.QUADS,
    RDF_TO_JELLY_TESTS_DIR / PhysicalTypeTestCasesDir.GRAPHS,
    glob="neg_*",
)
def test_serializing_fails(path: Path) -> None:
    options_filename = path / "stream_options.jelly"
    test_id = id_from_path(path)
    actual_out = TEST_OUTPUTS_DIR / f"{test_id}.jelly"
    with pytest.raises(Exception):  # TODO: more specific  # noqa: PT011,B017,TD002
        write_graph_or_dataset(
            *map(str, path.glob("in_*")),
            options=options_filename,
            out_filename=actual_out,
        )


@walk_directories(
    RDF_TO_JELLY_TESTS_DIR / PhysicalTypeTestCasesDir.TRIPLES,
    RDF_TO_JELLY_TESTS_DIR / PhysicalTypeTestCasesDir.QUADS,
    RDF_TO_JELLY_TESTS_DIR / PhysicalTypeTestCasesDir.QUADS,
    glob="neg_*",
)
@needs_jelly_cli
def test_generic_serializing_fails(path: Path) -> None:
    if path:
        options_filename = path / "stream_options.jelly"
        test_id = id_from_path(path)
        actual_out = TEST_OUTPUTS_DIR / f"{test_id}.jelly"
        with pytest.raises(Exception):  # TODO: more specific  # noqa: PT011,B017,TD002
            write_generic_sink(
                *map(str, path.glob("in_*")),
                options=options_filename,
                out_filename=actual_out,
            )



from __future__ import annotations

from collections.abc import Iterable, Sequence
from enum import Enum
from typing import Any, ClassVar, TypeVar
from typing_extensions import TypeAlias

from pyjelly import jelly, options
from pyjelly.errors import JellyConformanceError
from pyjelly.serialize.lookup import LookupEncoder


def split_iri(iri_string: str) -> tuple[str, str]:
    """
    Split iri into prefix and name.

    Args:
        iri_string (str): full iri string.

    Returns:
        tuple[str, str]: iri's prefix and name.

    """
    name = iri_string
    prefix = ""
    for sep in "#", "/":
        prefix, char, name = iri_string.rpartition(sep)
        if char:
            return prefix + char, name
    return prefix, name


T = TypeVar("T")
RowsAnd: TypeAlias = tuple[Sequence[jelly.RdfStreamRow], T]
RowsAndTerm: TypeAlias = "RowsAnd[jelly.RdfIri | jelly.RdfLiteral | str | \
    jelly.RdfDefaultGraph | jelly.RdfTriple]"


class TermEncoder:
    TERM_ONEOF_NAMES: ClassVar = {
        jelly.RdfIri: "iri",
        jelly.RdfLiteral: "literal",
        str: "bnode",
        jelly.RdfDefaultGraph: "default_graph",
        jelly.RdfTriple: "triple_term",
    }

    def __init__(
        self,
        lookup_preset: options.LookupPreset | None = None,
    ) -> None:
        if lookup_preset is None:
            lookup_preset = options.LookupPreset()
        self.lookup_preset = lookup_preset
        self.names = LookupEncoder(lookup_size=lookup_preset.max_names)
        self.prefixes = LookupEncoder(lookup_size=lookup_preset.max_prefixes)
        self.datatypes = LookupEncoder(lookup_size=lookup_preset.max_datatypes)

    def encode_iri(self, iri_string: str) -> RowsAnd[jelly.RdfIri]:
        """
        Encode iri.

        Args:
            iri_string (str): full iri in string format.

        Returns:
            RowsAnd[jelly.RdfIri]: extra rows and protobuf RdfIri message.

        """
        prefix, name = split_iri(iri_string)
        if self.prefixes.lookup.max_size:
            prefix_entry_index = self.prefixes.encode_entry_index(prefix)
        else:
            name = iri_string
            prefix_entry_index = None

        name_entry_index = self.names.encode_entry_index(name)
        term_rows = []

        if prefix_entry_index is not None:
            prefix_entry = jelly.RdfPrefixEntry(id=prefix_entry_index, value=prefix)
            term_rows.append(jelly.RdfStreamRow(prefix=prefix_entry))

        if name_entry_index is not None:
            name_entry = jelly.RdfNameEntry(id=name_entry_index, value=name)
            term_rows.append(jelly.RdfStreamRow(name=name_entry))

        prefix_index = self.prefixes.encode_prefix_term_index(prefix)
        name_index = self.names.encode_name_term_index(name)
        return term_rows, jelly.RdfIri(prefix_id=prefix_index, name_id=name_index)

    def encode_default_graph(self) -> RowsAnd[jelly.RdfDefaultGraph]:
        """
        Encode default graph.

        Returns:
            RowsAnd[jelly.RdfDefaultGraph]: empty extra rows and
                default graph message.

        """
        return (), jelly.RdfDefaultGraph()

    def encode_bnode(self, bnode: str) -> RowsAnd[str]:
        """
        Encode blank node (BN).

        Args:
            bnode (str): BN internal identifier in string format.

        Returns:
            RowsAnd[str]: empty extra rows and original BN string.

        """
        return (), bnode

    def encode_literal(
        self,
        *,
        lex: str,
        language: str | None = None,
        datatype: str | None = None,
    ) -> RowsAnd[jelly.RdfLiteral]:
        """
        Encode literal.

        Args:
            lex (str): lexical form/literal value
            language (str | None, optional): langtag. Defaults to None.
            datatype (str | None, optional): data type if
            it is a typed literal. Defaults to None.

        Raises:
            JellyConformanceError: if datatype specified while
                datatable is not used.

        Returns:
            RowsAnd[jelly.RdfLiteral]: extra rows (i.e., datatype entries)
                and RdfLiteral message.

        """
        datatype_id = None
        term_rows: tuple[()] | tuple[jelly.RdfStreamRow] = ()

        if datatype and datatype != options.STRING_DATATYPE_IRI:
            if self.datatypes.lookup.max_size == 0:
                msg = (
                    f"can't encode literal with type {datatype}: "
                    "datatype lookup cannot be used if disabled "
                    "(its size was set to 0)"
                )
                raise JellyConformanceError(msg)
            datatype_entry_id = self.datatypes.encode_entry_index(datatype)

            if datatype_entry_id is not None:
                entry = jelly.RdfDatatypeEntry(id=datatype_entry_id, value=datatype)
                term_rows = (jelly.RdfStreamRow(datatype=entry),)

            datatype_id = self.datatypes.encode_datatype_term_index(datatype)

        return term_rows, jelly.RdfLiteral(
            lex=lex,
            langtag=language,
            datatype=datatype_id,
        )

    def encode_quoted_triple(self, terms: Iterable[object]) -> RowsAndTerm:
        """
        Encode a quoted triple.

        Notes:
            Although a triple, it is treated as a part of a statement.
            Repeated terms are not used when encoding quoted triples.

        Args:
            terms (Iterable[object]): triple terms to encode.

        Returns:
            RowsAndTerm: additional stream rows with preceeding
                information (prefixes, names, datatypes rows, if any)
                and the encoded triple row.

        """
        statement: dict[str, Any] = {}
        rows: list[jelly.RdfStreamRow] = []
        for slot, term in zip(Slot, terms):
            extra_rows, value = self.encode_any(term, slot)
            oneof = self.TERM_ONEOF_NAMES[type(value)]
            rows.extend(extra_rows)
            field = f"{slot}_{oneof}"
            statement[field] = value
        return rows, jelly.RdfTriple(**statement)

    def encode_any(self, term: object, slot: Slot) -> RowsAndTerm:
        msg = f"unsupported term type: {type(term)}"
        raise NotImplementedError(msg)


class Slot(str, Enum):
    """Slots for encoding RDF terms."""

    subject = "s"
    predicate = "p"
    object = "o"
    graph = "g"

    def __str__(self) -> str:
        return self.value


def encode_statement(
    terms: Iterable[object],
    term_encoder: TermEncoder,
    repeated_terms: dict[Slot, object],
) -> tuple[list[jelly.RdfStreamRow], dict[str, Any]]:
    """
    Encode a statement.

    Args:
        terms (Iterable[object]): original terms to encode
        term_encoder (TermEncoder): encoder with lookup tables
        repeated_terms (dict[Slot, object]): dictionary of repeated terms.

    Returns:
        tuple[list[jelly.RdfStreamRow], dict[str, Any]]:
            extra rows to append and jelly terms.

    """
    statement: dict[str, object] = {}
    rows: list[jelly.RdfStreamRow] = []
    for slot, term in zip(Slot, terms):
        if repeated_terms[slot] != term:
            extra_rows, value = term_encoder.encode_any(term, slot)
            oneof = term_encoder.TERM_ONEOF_NAMES[type(value)]
            rows.extend(extra_rows)
            field = f"{slot}_{oneof}"
            statement[field] = value
            repeated_terms[slot] = term
    return rows, statement


def encode_triple(
    terms: Iterable[object],
    term_encoder: TermEncoder,
    repeated_terms: dict[Slot, object],
) -> list[jelly.RdfStreamRow]:
    """
    Encode one triple.

    Args:
        terms (Iterable[object]): original terms to encode
        term_encoder (TermEncoder): current encoder with lookup tables
        repeated_terms (dict[Slot, object]): dictionary of repeated terms.

    Returns:
        list[jelly.RdfStreamRow]: list of rows to add to the current flow.

    """
    rows, statement = encode_statement(terms, term_encoder, repeated_terms)
    row = jelly.RdfStreamRow(triple=jelly.RdfTriple(**statement))
    rows.append(row)
    return rows


def encode_quad(
    terms: Iterable[object],
    term_encoder: TermEncoder,
    repeated_terms: dict[Slot, object],
) -> list[jelly.RdfStreamRow]:
    """
    Encode one quad.

    Args:
        terms (Iterable[object]): original terms to encode
        term_encoder (TermEncoder): current encoder with lookup tables
        repeated_terms (dict[Slot, object]): dictionary of repeated terms.

    Returns:
        list[jelly.RdfStreamRow]: list of messages to append to current flow.

    """
    rows, statement = encode_statement(terms, term_encoder, repeated_terms)
    row = jelly.RdfStreamRow(quad=jelly.RdfQuad(**statement))
    rows.append(row)
    return rows


def encode_namespace_declaration(
    name: str,
    value: str,
    term_encoder: TermEncoder,
) -> list[jelly.RdfStreamRow]:
    """
    Encode namespace declaration.

    Args:
        name (str): namespace prefix label
        value (str): namespace iri
        term_encoder (TermEncoder): current encoder

    Returns:
        list[jelly.RdfStreamRow]: list of messages to append to current flow.

    """
    [*rows], iri = term_encoder.encode_iri(value)
    declaration = jelly.RdfNamespaceDeclaration(name=name, value=iri)
    row = jelly.RdfStreamRow(namespace=declaration)
    rows.append(row)
    return rows


def encode_options(
    lookup_preset: options.LookupPreset,
    stream_types: options.StreamTypes,
    params: options.StreamParameters,
) -> jelly.RdfStreamRow:
    """
    Encode stream options to ProtoBuf message.

    Args:
        lookup_preset (options.LookupPreset): lookup tables options
        stream_types (options.StreamTypes): physical and logical types
        params (options.StreamParameters): other params.

    Returns:
        jelly.RdfStreamRow: encoded stream options row

    """
    return jelly.RdfStreamRow(
        options=jelly.RdfStreamOptions(
            stream_name=params.stream_name,
            physical_type=stream_types.physical_type,
            generalized_statements=params.generalized_statements,
            rdf_star=params.rdf_star,
            max_name_table_size=lookup_preset.max_names,
            max_prefix_table_size=lookup_preset.max_prefixes,
            max_datatype_table_size=lookup_preset.max_datatypes,
            logical_type=stream_types.logical_type,
            version=params.version,
        )
    )



from abc import ABC, abstractmethod
from typing import Union

from rdflib import Dataset, Graph

from pyjelly.options import LookupPreset

TripleGraphType = Union[Graph]
QuadGraphType = Union[Dataset]


class BaseSerDes(ABC):
    """
    Base class for serialization and deserialization tests.

    This class provides a common interface for reading and writing
    jelly/nquads/ntriples from and to byte streams. It is intended to be subclassed
    for specific libraries used for serialization and deserialization.

    Attributes:
        name (str): The name of the serialization/deserialization library.

    """

    name: str

    def __init__(self, name: str) -> None:
        self.name = name

    @abstractmethod
    def read_quads(self, in_bytes: bytes) -> QuadGraphType:
        """
        Read quads from bytes to a graph-like structure.

        Args:
            in_bytes (bytes): The input bytes to read from.

        Returns:
            QuadGraphType: The graph-like structure containing the quads.

        """

    @abstractmethod
    def write_quads(self, in_graph: QuadGraphType) -> bytes:
        """
        Write serialized quads to bytes.

        Args:
            in_graph (QuadGraphType): The graph-like structure containing the quads.

        Returns:
            bytes: The output bytes containing the serialized quads.

        """

    @abstractmethod
    def read_quads_jelly(self, in_bytes: bytes) -> QuadGraphType:
        """
        Read quads from jelly bytes.

        Args:
            in_bytes (bytes): The input bytes to read from.

        Returns:
            QuadGraphType: The graph-like structure containing the quads.

        """

    @abstractmethod
    def write_quads_jelly(
        self, in_graph: QuadGraphType, preset: LookupPreset, frame_size: int
    ) -> bytes:
        """
        Write quads to jelly bytes.

        Args:
            in_graph (QuadGraphType): The graph-like structure containing the quads.
            preset (LookupPreset): The stream options lookup preset for serialization.
            frame_size (int): The size of the frame for serialization.

        Returns:
            bytes: The output bytes containing the serialized quads.

        """

    @abstractmethod
    def read_triples(self, in_bytes: bytes) -> TripleGraphType:
        """
        Read triples from bytes to a graph-like structure.

        Args:
            in_bytes (bytes): The input bytes to read from.

        Returns:
            TripleGraphType: The graph-like structure containing the triples.

        """

    @abstractmethod
    def write_triples(self, in_graph: TripleGraphType) -> bytes:
        """
        Write triples to bytes.

        Args:
            in_graph (TripleGraphType): The graph-like structure containing the triples.

        Returns:
            bytes: The output bytes containing the serialized triples.

        """

    @abstractmethod
    def read_triples_jelly(self, in_bytes: bytes) -> TripleGraphType:
        """

        Read triples from jelly bytes.

        Args:
            in_bytes (bytes): The input bytes to read from.

        Returns:
            TripleGraphType: The graph-like structure containing the triples.

        """

    @abstractmethod
    def write_triples_jelly(
        self, in_graph: TripleGraphType, preset: LookupPreset, frame_size: int
    ) -> bytes:
        """
        Write triples to jelly bytes.

        Args:
            in_graph (TripleGraphType): The graph-like structure containing the triples.
            preset (LookupPreset): The stream options lookup preset for serialization.
            frame_size (int): The size of the frame for serialization.

        Returns:
            bytes: The output bytes containing the serialized triples.

        """
